from __future__ import division, print_function, unicode_literals

import random
import types
import unittest

import numpy.testing

try:
    import unittest.mock as mock
except ImportError:
    import mock
from pyoma.browser.db import *
from pyoma.browser import tablefmt
from pyoma.browser.convert import _load_taxonomy_without_ref_to_itselfs as load_taxtable_with_defaults
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio import SeqIO


class TestHelperFunctions(unittest.TestCase):
    def test_counter(self):
        self.assertEqual(0, count_elements([]))
        self.assertEqual(3, count_elements("abc"))
        recarray = numpy.zeros(2, dtype=[("A", "i4"), ("B", "f8")])
        self.assertEqual(2, count_elements(recarray))


def find_path_to_test_db(dbfn="TestDb.h5"):
    """We try to load the dbfn first from the same directory, and afterwards from
    the path given by the PYOMA_DB_PATH environment variable.

    :returns: path to database
    :rtype: str
    :raises IOError: if db does not exist."""

    def is_hdf5_file(fname):
        with open(fname, "rb") as fh:
            return fh.read(4) == b"\x89HDF"

    path1 = os.path.join(os.path.dirname(__file__), dbfn)
    if os.path.isfile(path1) and is_hdf5_file(path1):
        return path1
    path2 = os.path.abspath(os.path.join(os.getenv("PYOMA_DB_PATH", "./"), dbfn))
    if os.path.isfile(path2) and is_hdf5_file(path2):
        return path2
    else:
        raise IOError("cannot access {}. (Tried {} and {})".format(dbfn, path1, path2))


class TestWithDbInstance(unittest.TestCase):
    db = None
    db_file_name = "TestDb.h5"

    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db(cls.db_file_name)
        logger.info("Loading %s for DatabaseTests", path)
        cls.db = Database(path)

    @classmethod
    def tearDownClass(cls):
        cls.db.close()


class DatabaseTests(TestWithDbInstance):
    def test_get_vpairs_of_entry_with_orthologs(self):
        for entry_nr, exp_vps_cnt in [(12, 3), (1, 0), (4, 1)]:
            vps = self.db.get_vpairs(entry_nr)
            self.assertTrue(isinstance(vps, numpy.ndarray))
            self.assertEqual(exp_vps_cnt, len(vps))
            self.assertEqual(
                sorted(["EntryNr1", "EntryNr2", "RelType", "Distance", "Score"]),
                sorted(vps.dtype.fields.keys()),
            )

    def test_get_vpairs_between_species_is_same_as_loading_all_vpairs(self):
        g1, g2 = "YEAST", "SCHPO"
        t0 = time.time()
        rng1 = self.db.id_mapper["OMA"].genome_range(g1)
        rng2 = self.db.id_mapper["OMA"].genome_range(g2)
        expected = []
        for en in range(rng1[0], rng1[1] + 1):
            for vp in self.db.get_vpairs(en):
                if rng2[0] <= vp["EntryNr2"] <= rng2[1]:
                    expected.append(vp)
        expected = numpy.fromiter(expected, dtype=expected[0].dtype)
        t1 = time.time()
        actual = self.db.get_vpairs_between_species_pair(g1, g2)
        t2 = time.time()
        numpy.testing.assert_equal(expected, actual)
        self.assertLess(10 * (t2 - t1), t1 - t0, "expect at least 10x speedup")

    def test_get_hoginducedpairs_between_species_is_same_as_loading_all_hoginduced(
        self,
    ):
        g1, g2 = "YEAST", "SCHPO"
        t0 = time.time()
        rng1 = self.db.id_mapper["OMA"].genome_range(g1)
        rng2 = self.db.id_mapper["OMA"].genome_range(g2)
        expected = []
        for en in range(rng1[0], rng1[1] + 1):
            for vp in self.db.get_hog_induced_pairwise_orthologs(en):
                if rng2[0] <= vp["EntryNr"] <= rng2[1]:
                    expected.append((en, int(vp["EntryNr"])))
        expected = numpy.fromiter(expected, dtype=[("EntryNr1", "u4"), ("EntryNr2", "u4")])
        t1 = time.time()
        actual = self.db.get_hog_induced_orthologs_between_genome_pair(g1, g2)
        actual = actual[["EntryNr1", "EntryNr2"]].copy()
        actual.sort()
        t2 = time.time()
        numpy.testing.assert_equal(expected, actual)
        self.assertLess(5 * (t2 - t1), t1 - t0, "expect at least 5x speedup")

    def test_neighborhood_close_to_boundary(self):
        query, window = 3, 7
        neighbors, idx = self.db.neighbour_genes(query, window)
        self.assertEqual(query - 1, idx)
        self.assertEqual(neighbors["EntryNr"][idx], query)
        expected_entry_nrs = numpy.arange(1, query + window + 1, dtype="i4")
        self.assertTrue(numpy.array_equal(expected_entry_nrs, neighbors["EntryNr"]))

    def test_count_rows_of_index_column_with_value(self):
        entrynr = 12
        vps = self.db.get_vpairs(entrynr)
        vp_tab = self.db._get_vptab(entrynr)
        self.assertEqual(len(vps), count_rows_of_index_column_with_value(vp_tab, "EntryNr1", entrynr))

        # check that it raises KeyError if operated on a column that is not indexed
        with self.assertRaises(KeyError, msg="Operating on non-indexed column should raise KeyError"):
            count_rows_of_index_column_with_value(vp_tab, "EntryNr2", entrynr)

    def test_hog_family(self):
        entry = numpy.zeros(1, dtype=tables.dtype_from_descr(tablefmt.ProteinTable))
        entry["OmaHOG"] = b""
        with self.assertRaises(Singleton):
            self.db.hog_family(entry[0])
        entry["OmaHOG"] = b"HOG:0000523"
        self.assertEqual(523, self.db.hog_family(entry[0]))

    def test_orthoxml(self):
        xml = self.db.get_orthoxml(33).decode()
        # we simply check that orthoxml starts with <?xml and ends with an orthoxml tag
        self.assertTrue(xml.startswith("<?xml "))
        self.assertTrue(xml.endswith("</orthoXML>\n"))

    def test_hog_lex_range(self):
        cases = [
            (b"HOG:001", (b"HOG:001", b"HOG:002")),
            (b"HOG:001.1a.2b", (b"HOG:001.1a.2b", b"HOG:001.1a.2c")),
        ]
        for hog, rng in cases:
            self.assertEqual(rng, self.db._hog_lex_range(hog))

    def test_fam_member(self):
        memb = self.db.member_of_fam(1)
        self.assertEqual(2, len(memb))

    def test_hog_members(self):
        cases = [("Eukaryota", 5), ("Fungi", 3), ("Taphrinomycotina", 0)]
        for level, exp_member_cnt in cases:
            if exp_member_cnt == 0:
                with self.assertRaises(ValueError):
                    self.db.hog_members(12, level)
            else:
                self.assertEqual(len(self.db.hog_members(12, level)), exp_member_cnt)

    def test_hogids_at_level(self):
        cases = [
            [(2, "Ascomycota"), numpy.array([b"HOG:0000002"])],
            [
                (2, "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)"),
                numpy.array([b"HOG:0000002.2a", b"HOG:0000002.2b"]),
            ],
        ]

        for case in cases:
            args, expected = case
            levels = self.db.get_subhogids_at_level(*args)
            self.assertTrue(
                numpy.array_equal(expected, levels),
                "test of tes_hogids_at_level failed for {}: {}".format(args, levels),
            )

    def test_get_hog_without_level(self):
        for case in ("HOG:0000002.2b", "HOG:0000001"):
            with self.subTest(case=case):
                hog = self.db.get_hog(case)
                self.assertTrue(hog["IsRoot"])
                self.assertEqual(case, hog["ID"].decode())

    def test_get_parent_hogs(self):
        for hog_id, level in (
            ("HOG:0000474.1d", "Ascomycota"),
            ("HOG:0000165", "Fungi"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hogs = self.db.get_parent_hogs(hog_id, level)
                try:
                    k = hog_id.index(".")
                    base_id = hog_id[:k]
                except ValueError:
                    base_id = hog_id
                self.assertEqual(base_id, hogs[0].hog_id)
                self.assertTrue(hogs[0].is_root)
                self.assertEqual(hog_id, hogs[-1].hog_id)
                self.assertEqual(level, hogs[-1].level)

    def test_get_subhogs(self):
        for hog_id, level in (
            ("HOG:0000474.1d", "Ascomycota"),
            ("HOG:0000165", "Fungi"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hogs = list(self.db.get_subhogs(hog_id, level))
                self.assertEqual(hogs[0].hog_id, hog_id)
                self.assertNotIn(level, [h.level for h in hogs])

    def test_get_subhogs_with_subids(self):
        for hog_id, level, exp_subhogid in (
            ("HOG:0000474", "Eukaryota", "HOG:0000474.1d.2c"),
            ("HOG:0000165", "Fungi", "HOG:0000165.1a"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hogs = list(self.db.get_subhogs(hog_id, level=level, include_subids=True))
                self.assertIn(exp_subhogid, [h.hog_id for h in hogs])
                self.assertNotIn(level, [h.level for h in hogs])

    def test_get_hog_with_level(self):
        for hog_id, level in (
            ("HOG:0000002.2b", "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)"),
            ("HOG:0000005.1b", "Saccharomycetaceae"),
            ("HOG:0000005", "Fungi"),
        ):
            with self.subTest(hog_id=hog_id, level=level):
                hog = self.db.get_hog(hog_id, level=level)
                self.assertEqual(hog_id, hog["ID"].decode())
                self.assertEqual(level, hog["Level"].decode())

    def test_member_of_hog_id(self):
        cases = [
            [("HOG:0000082.1b", None), 2],
            [("HOG:0000082.1b", "Saccharomycetaceae"), 2],
            [
                (
                    "HOG:0000082.1b",
                    "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                ),
                1,
            ],
            [("HOG:0000082.1a", "Mammalia"), 0],
        ]
        for args, expected_len in cases:
            members = self.db.member_of_hog_id(*args)
            self.assertEqual(len(members), expected_len)

    def test_sorted_genomes(self):
        for root in ("YEAST", "ASHGO"):
            order = self.db.id_mapper["OMA"].species_ordering(root)
            self.assertEqual(
                order[root],
                0,
                "{} should be first genome, but comes at {}".format(root, order[root]),
            )

    def test_main_isoform_from_species_id(self):
        query = "PLAF7"
        rng = self.db.id_mapper["OMA"].genome_range(query)
        mains = self.db.main_isoforms(query)
        self.assertGreaterEqual(rng[1] - rng[0] + 1, len(mains))

    def test_valid_protein_seq_check(self):
        for c in (b"ADRIAN", "adrian", b"XaMGAt"):
            self.assertTrue(
                self.db.seq_search.contains_only_valid_chars(c),
                "'{}' should be a valid protein seq".format(c),
            )
        for c in (b"XBra", "T5HSE", "xxnnebqpynn", " "):
            self.assertFalse(
                self.db.seq_search.contains_only_valid_chars(c),
                "'{}' should be an invalid protein seq".format(c),
            )

    def test_exact_search(self):
        # Test for 10 random
        for _ in range(10):
            i = random.randint(0, len(self.db.db.root.Protein.Entries))
            enr = i + 1
            s = self.db.get_sequence(enr)
            self.assertTrue(
                (enr in set(self.db.seq_search.search(s, is_sanitised=True)[1])),
                "exact search for entry {} failed.".format(i),
            )

    def get_random_subsequence(self, minlen=10):
        i = random.randint(0, len(self.db.db.root.Protein.Entries))
        elen = self.db.db.root.Protein.Entries[i]["SeqBufferLength"] - 1
        enr = i + 1

        ii = random.randint(0, elen - minlen)
        jj = random.randint(ii + minlen, elen)

        s = self.db.get_sequence(enr)[ii:jj]
        return s, enr, ii, jj

    def test_approx_search(self):
        # Test for random subsequence of 10 random sequences.
        for _ in range(10):
            s, enr, start_idx, end_idx = self.get_random_subsequence()
            approx_search_results = self.db.seq_search.approx_search(s, is_sanitised=True)
            self.assertIn(
                enr,
                {z[0] for z in approx_search_results},
                "approx search for entry {}[{}:{}] failed.".format(enr - 1, start_idx, end_idx),
            )

    def test_specific_approx_search_that_failed_on_jenkins(self):
        # TODO fix this problem!
        enrs = [15885, 16452]
        ranges = [(39, 376), (55, 140)]
        for enr, (start_idx, end_idx) in zip(enrs, ranges):
            seq = self.db.get_sequence(enr)[start_idx:end_idx]
            approx_search_results = self.db.seq_search.approx_search(seq, is_sanitised=True)
            enrs_with_approx_match = {z[0] for z in approx_search_results}
            self.assertIn(enr, enrs_with_approx_match)

    def test_map_to_hog(self):
        hog_mapper = ClosestSeqMapper(self.db)
        seqs = []
        entry_nrs = []
        for _ in range(10):
            s, enr, start_idx, end_idx = self.get_random_subsequence()
            entry_nrs.append(enr)
            seqs.append(
                SeqRecord(
                    Seq(s.decode()),
                    id=str(enr),
                    annotations={"molecule_type": "protein"},
                )
            )
        res = list(
            filter(
                lambda x: x.query == str(x.closest_entry_nr),
                hog_mapper.imap_sequences(seqs),
            )
        )
        self.assertEqual(10, len(res))
        for case in res:
            self.assertTrue(0 < case.distance < 1)

    def test_oma_group_from_numeric_id(self):
        group_id = 5
        grp = self.db.oma_group_members(group_id)
        self.assertEqual(4, len(grp))
        for e in grp:
            self.assertEqual(group_id, e["OmaGroup"])

    def test_fingerprint(self):
        fingerprint, grp_nr = "ESRTELL", 2617
        grp = self.db.oma_group_members(fingerprint)
        self.assertLessEqual(2, len(grp))
        for e in grp:
            self.assertEqual(grp_nr, e["OmaGroup"])

    def test_exon_structure(self):
        query = 14677  # Q8I237
        exons = self.db.get_exons(query)
        self.assertEqual(3, len(exons))

    def test_go_term_search(self):
        query = "GO:0004575"
        nrs = self.db.entrynrs_with_go_annotation(query, evidence="IDA")
        self.assertGreaterEqual(len(nrs), 1, "query GO term is known to occure at least in MAL32_YEAST")
        for enr in nrs:
            self.assertIn(4575, self.db.get_gene_ontology_annotations(enr)["TermNr"])

    def test_go_ic_values(self):
        terms = ["GO:003674", "GO:008150"]
        for term in terms:
            with self.subTest(case=term):
                gof = self.db.freq_aware_gene_ontology
                ic = gof.ic(term)
                self.assertAlmostEqual(0, ic)

    def test_go_lin_same_term_is_1(self):
        self.assertEqual(1, self.db.freq_aware_gene_ontology.lin_similarity(55, 55))

    def test_go_semantic_sim_of_same_term(self):
        self.assertGreater(1, self.db.freq_aware_gene_ontology.semantic_similarity(55, 55))

    def test_mindepth_of_go(self):
        examples = [(8150, 0), (168, 7), (5689, 5)]
        for term, exp_depth in examples:
            with self.subTest(case=term):
                self.assertEqual(exp_depth, self.db.gene_ontology.ensure_term(term).min_depth)

    def test_induced_pairwise_orthologs(self):
        query = "YEAST3523"
        query_entry = self.db.ensure_entry(self.db.id_resolver.resolve(query))
        orthologs = self.db.get_hog_induced_pairwise_orthologs(query_entry)
        self.assertEqual(3, len(orthologs))
        self.assertCountEqual([b"1:1", b"1:1", b"m:1"], orthologs["RelType"])

    def test_induced_pairwise_paralogs(self):
        query = "YEAST12"
        query_entry = self.db.ensure_entry(self.db.id_resolver.resolve(query))
        orthologs = self.db.get_hog_induced_pairwise_paralogs(query_entry)
        self.assertEqual(1, len(orthologs))
        self.assertEqual(
            b"Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
            orthologs["DivergenceLevel"],
        )

    def test_ancestral_synteny_graph(self):
        query_hog = "HOG:0000005"
        query_level = "Fungi"
        graph = self.db.get_syntentic_hogs(query_level, query_hog, steps=2)
        self.assertIn(query_hog, graph)

    def test_ancestral_syteny_of_genome(self):
        query_level = "Fungi"
        graph = self.db.get_syntenic_hogs(level=query_level)
        self.assertIn("HOG:0000005", graph)
        self.assertEqual(len(graph.nodes), self.db.count_hogs_at_level(query_level))

    def test_extant_syntey_of_genome(self):
        query_level = "YEAST"
        graph = self.db.get_extant_synteny_graph(query_level)
        self.assertIn("YEAST00012", graph)


class XRefDatabaseMock(Database):
    def __init__(self, name=None):
        if name is None:
            name = "xref.h5"
        f = tables.open_file(name, "w", driver="H5FD_CORE", driver_core_backing_store=0)
        entries = numpy.zeros(5, tables.dtype_from_descr(tablefmt.ProteinTable))
        entries["EntryNr"] = numpy.arange(1, 6)
        t = f.create_table("/Protein", "Entries", obj=entries, createparents=True)
        t.colinstances["EntryNr"].create_csindex()
        xref = numpy.zeros(10, tables.dtype_from_descr(tablefmt.XRefTable))
        xref["EntryNr"] = numpy.arange(1, 6, 0.5).astype(numpy.int32)
        xref["XRefSource"] = numpy.tile([0, 20], 5)
        xref["XRefId"] = ["XA{:03}g1.4".format(i) for i in range(10)]
        xref["Verification"] = tuple(itertools.islice(itertools.cycle([0, 2, 4]), 10))
        x = f.create_table("/", "XRef", tablefmt.XRefTable, obj=xref)
        x.colinstances["EntryNr"].create_csindex()
        x.colinstances["XRefId"].create_csindex()
        f.create_group("/", "XRef_Index")
        for n in ("suffix", "buffer", "offset"):
            f.create_carray("/XRef_Index", n, obj=numpy.ones((5,), "i4"))
        self.db = f
        self._on_close_notify = []
        self._close_fh = True


class XRefIdMapperTest(unittest.TestCase):
    @classmethod
    def setUp(self):
        patch_db = XRefDatabaseMock()
        self.xrefmapper = XrefIdMapper(patch_db)

    def tearDown(self):
        self.xrefmapper._db.close()

    def test_multiple_xrefs_per_entry(self):
        xref_e1 = self.xrefmapper.map_entry_nr(1)
        self.assertEqual(len(xref_e1), 2)

    def test_map_many_entries(self):
        all_mapped = self.xrefmapper.map_many_entry_nrs(numpy.arange(1, 4))
        expected_len = 4 if isinstance(self.xrefmapper, XRefNoApproximateIdMapper) else 6
        self.assertEqual((expected_len,), all_mapped.shape)
        self.assertEqual(self.xrefmapper.xref_tab.dtype, all_mapped.dtype)

    def test_entry_nr_range(self):
        all_mapped = self.xrefmapper.map_many_entry_nrs(numpy.arange(1, 4))
        range_mapped = self.xrefmapper.map_entry_nr_range(1, 4)
        numpy.testing.assert_equal(all_mapped, range_mapped)

    def test_map_entry_iterator(self):
        it = self.xrefmapper.iter_xrefs_for_entry_nr(1)
        self.assertTrue(isinstance(it, types.GeneratorType), "not a generator")
        exp_xrefs = ["XA000g1.4", "XA001g1.4"]
        for dic in it:
            self.assertIn(dic["xref"], exp_xrefs)

    def test_search_of_modified_xref(self):
        xref = "XA002g1.4"
        res = self.xrefmapper.search_xref(xref)
        self.assertEqual(2, res["EntryNr"])

    def test_verification_is_returned(self):
        res = self.xrefmapper.map_entry_nr(2)
        if isinstance(self.xrefmapper, XRefNoApproximateIdMapper):
            expected = ["exact"]
        else:
            expected = ["modified", "exact"]
        self.assertEqual(expected, [z["seq_match"] for z in res])


class XRefNoApproxIdMapperTest(XRefIdMapperTest):
    def setUp(self):
        patch_db = XRefDatabaseMock()
        self.xrefmapper = XRefNoApproximateIdMapper(patch_db)

    def test_modified_xref_not_returned_in_map(self):
        res = self.xrefmapper.map_entry_nr(2)
        xrefs = [x["xref"] for x in res]
        self.assertNotIn("XA002g1.4", xrefs)
        self.assertIn("XA003g1.4", xrefs)


class IdResolverTests(unittest.TestCase):
    @classmethod
    def setUp(self):
        patch_db = XRefDatabaseMock("test_idresolver.h5")
        self.xrefmapper = XrefIdMapper(patch_db)
        self.id_resolver = IDResolver(patch_db)
        patch_db.id_mapper = {"XRef": self.xrefmapper}

    def tearDown(self):
        self.xrefmapper._db.close()

    def test_search_of_modified_xref(self):
        xref = "XA002g1.4"
        res = self.id_resolver.search_xrefs(xref, return_seq_modified=True)
        self.assertEqual(2, res[0])
        self.assertTrue(res[1], "expected that {} is a modified sequence".format(xref))

    def test_search_of_unchecked_xref(self):
        xref = "XA001g1.4"
        res = self.id_resolver.search_xrefs(xref, return_seq_modified=True)
        self.assertEqual(1, res[0])
        self.assertFalse(res[1], "expected that {} is a unchecked sequence".format(xref))


class TaxonomyTest(unittest.TestCase):
    tax_input = None
    maxDiff = None  # show complete strings if test fails

    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db("TestDb.h5")
        h5 = tables.open_file(path)
        cls.tax_input = h5.root.Taxonomy.read()
        h5.close()

    def setUp(self):
        self.tax = Taxonomy(self.tax_input)

    def test_parents(self):
        lin = [x["Name"].decode() for x in self.tax.get_parent_taxa(284811)]
        self.assertEqual(
            lin,
            [
                "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                "Eremothecium",
                "Saccharomycetaceae",
                "Saccharomycetales",
                "Saccharomycetes",
                "Saccharomycotina",
                "saccharomyceta",
                "Ascomycota",
                "Dikarya",
                "Fungi",
                "Opisthokonta",
                "Eukaryota",
            ],
        )

    def test_newick(self):
        member = frozenset([self.tax._taxon_from_numeric(x)["Name"] for x in self.tax.tax_table["NCBITaxonId"]])
        phylo = self.tax.get_induced_taxonomy(member, collapse=True)
        expected = "(((Ashbya gossypii [strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056],Saccharomyces cerevisiae [strain ATCC 204508 / S288c])Saccharomycetaceae,Schizosaccharomyces pombe [strain 972 / ATCC 24843])Ascomycota,Plasmodium falciparum [isolate 3D7])Eukaryota;"
        expected = expected.replace(" ", "_")
        self.assertEqual(expected, phylo.newick())

    def test_phylogeny(self):
        member = frozenset([self.tax._taxon_from_numeric(x)["Name"] for x in self.tax.tax_table["NCBITaxonId"]])
        phylo = self.tax.get_induced_taxonomy(member, collapse=True)
        expected = {
            "id": 2759,
            "name": "Eukaryota",
            "children": [
                {
                    "id": 4890,
                    "name": "Ascomycota",
                    "children": [
                        {
                            "id": 4893,
                            "name": "Saccharomycetaceae",
                            "children": [
                                {
                                    "id": 284811,
                                    "name": "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                                },
                                {
                                    "id": 559292,
                                    "name": "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                                },
                            ],
                        },
                        {
                            "id": 284812,
                            "name": "Schizosaccharomyces pombe (strain 972 / ATCC 24843)",
                        },
                    ],
                },
                {"id": 36329, "name": "Plasmodium falciparum (isolate 3D7)"},
            ],
        }
        self.assertEqual(expected, phylo.as_dict())

    def test_induced_tax_simple_subtree(self):
        members = [559292, 284811]
        phylo = self.tax.get_induced_taxonomy(members)
        expected = {
            "id": 0,
            "name": "LUCA",
            "children": [
                {
                    "id": 284811,
                    "name": "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                },
                {
                    "id": 559292,
                    "name": "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                },
            ],
        }
        self.assertEqual(expected, phylo.as_dict())

    def test_induced_tax_with_parents_subtree(self):
        members = [559292, 284811]
        phylo = self.tax.get_induced_taxonomy(members, augment_parents=True)
        expected = {
            "id": 4893,
            "name": "Saccharomycetaceae",
            "children": [
                {
                    "id": 284811,
                    "name": "Ashbya gossypii (strain ATCC 10895 / CBS 109.51 / FGSC 9923 / NRRL Y-1056)",
                },
                {
                    "id": 559292,
                    "name": "Saccharomyces cerevisiae (strain ATCC 204508 / S288c)",
                },
            ],
        }
        self.assertEqual(expected, phylo.as_dict())

    def test_subtree_tax(self):
        clade = "Fungi"
        subtax = self.tax.get_subtaxonomy_rooted_at(clade)
        self.assertIn(4893, subtax.tax_table["NCBITaxonId"])
        self.assertNotIn(36329, subtax.tax_table["NCBITaxonId"])

    def test_taxid_of_extent_genomes(self):
        taxids = set(self.tax.get_taxid_of_extent_genomes())
        self.assertEqual({284811, 559292, 284812, 36329}, taxids)

    def test_induced_tax_with_gaps(self):
        pass


class TaxonomyTestInternalLevelSpecies(unittest.TestCase):
    """This test asserts that species A, which is also an internal level
    of the taxonomy is still reported as an external species via a
    <sciname> (disambiguate <code>) leaf."""

    taxtab = numpy.array(
        [(10, 0, b"Root", 30), (20, 10, b"Outgroup", 25), (30, 10, b"A", 10), (40, 30, b"B", 5)],
        dtype=tables.dtype_from_descr(tablefmt.TaxonomyTable),
    )

    def setUp(self):
        patcher = mock.patch("pyoma.browser.models.Genome")
        self.addCleanup(patcher.stop)
        genome = patcher.start()
        type(genome).uniprot_species_code = mock.PropertyMock(return_value="HELLO")
        self.tax = Taxonomy(self.taxtab, genomes={20: genome, 30: genome, 40: genome})

    def test_newick(self):
        exp = "(Outgroup,(B,A_[disambiguate_HELLO])A)Root;"
        self.assertEqual(exp, self.tax.newick())

    def test_phyloxml(self):
        ins = (
            b"<scientific_name>A (disambiguate HELLO)</scientific_name>",
            b"<scientific_name>A</scientific_name>",
            b"<scientific_name>B</scientific_name>",
            b"<scientific_name>Outgroup</scientific_name>",
        )
        outs = (
            b"<scientific_name>B (disambiguate",
            b"<scientific_name>Outgroup (disambiguate",
        )
        res = self.tax.as_phyloxml()
        for part in ins:
            self.assertIn(part, res)
        for part in outs:
            self.assertNotIn(part, res)

    def test_dict_repr(self):
        res = self.tax.as_dict()
        self.assertDictEqual(
            {
                "name": "Root",
                "id": 10,
                "children": [
                    {"name": "Outgroup", "id": 20, "code": "HELLO"},
                    {
                        "name": "A",
                        "id": 30,
                        "children": [
                            {"name": "B", "id": 40, "code": "HELLO"},
                            {
                                "name": "A (disambiguate HELLO)",
                                "id": 30,
                                "code": "HELLO",
                            },
                        ],
                    },
                ],
            },
            res,
        )

    def test_induced_subtree_retains_internal_species(self):
        phylo = self.tax.get_induced_taxonomy([20, 40], augment_parents=True)
        self.assertIn(30, phylo.tax_table["NCBITaxonId"])


def _get_taxtab():
    # a random sample of 30 species from a production OMA release
    return load_taxtable_with_defaults(
        [
            (2, 0, b"Bacteria"),
            (633, 1649845, b"Yersinia pseudotuberculosis"),
            (1224, 2, b"Proteobacteria"),
            (1236, 1224, b"Gammaproteobacteria"),
            (2759, 0, b"Eukaryota"),
            (5204, 451864, b"Basidiomycota"),
            (5334, 5204, b"Schizophyllum commune"),
            (5660, 2759, b"Leishmania braziliensis"),
            (5786, 2759, b"Dictyostelium purpureum"),
            (5794, 2698737, b"Apicomplexa"),
            (5807, 5794, b"Cryptosporidium parvum"),
            (5864, 5794, b"Babesia"),
            (5866, 5864, b"Babesia bigemina"),
            (6412, 33317, b"Helobdella robusta"),
            (6960, 33317, b"Hexapoda"),
            (7263, 6960, b"Drosophila arizonae"),
            (9813, 32524, b"Procavia capensis"),
            (13735, 32524, b"Pelodiscus sinensis"),
            (28211, 1224, b"Alphaproteobacteria"),
            (29760, 2759, b"Vitis vinifera"),
            (32524, 33213, b"Amniota"),
            (33154, 2759, b"Opisthokonta"),
            (33213, 33154, b"Bilateria"),
            (33317, 33213, b"Protostomia"),
            (37360, 2698737, b"Plasmodiophora brassicae"),
            (55529, 2759, b"Guillardia theta"),
            (91347, 1236, b"Enterobacterales"),
            (119060, 1224, b"Burkholderiaceae"),
            (158441, 6960, b"Folsomia candida"),
            (164328, 2698737, b"Phytophthora ramorum"),
            (214092, 1649845, b"Yersinia pestis CO92"),
            (
                264198,
                119060,
                b"Cupriavidus pinatubonensis (strain JMP 134 / LMG 1197)",
            ),
            (
                273123,
                633,
                b"Yersinia pseudotuberculosis serotype I (strain IP32953)",
            ),
            (
                284812,
                451864,
                b"Schizosaccharomyces pombe (strain 972 / ATCC 24843)",
            ),
            (
                323261,
                1236,
                b"Nitrosococcus oceani (strain ATCC 19707 / BCRC 17464 / NCIMB 11848 / C-107)",
            ),
            (338187, 1236, b"Vibrio campbellii (strain ATCC BAA-1116 / BB120)"),
            (347255, 28211, b"Rickettsia africae (strain ESF-5)"),
            (353152, 5807, b"Cryptosporidium parvum (strain Iowa II)"),
            (
                355278,
                2,
                b"Leptospira biflexa serovar Patoc (strain Patoc 1 / Ames)",
            ),
            (391774, 1224, b"Desulfovibrio vulgaris subsp. vulgaris (strain DP4)"),
            (451864, 33154, b"Dikarya"),
            (
                502801,
                633,
                b"Yersinia pseudotuberculosis serotype IB (strain PB1/+)",
            ),
            (578458, 5334, b"Schizophyllum commune (strain H4-8 / FGSC 9210)"),
            (626418, 119060, b"Burkholderia glumae (strain BGR1)"),
            (640131, 91347, b"Klebsiella variicola (strain At-22)"),
            (
                713600,
                91347,
                b"Buchnera aphidicola subsp. Acyrthosiphon pisum (strain JF98)",
            ),
            (
                754035,
                28211,
                b"Mesorhizobium australicum (strain HAMBI 3006 / LMG 24608 / WSM2073)",
            ),
            (905079, 55529, b"Guillardia theta (strain CCMP2712)"),
            (1133968, 5864, b"Babesia microti (strain RI)"),
            (1230383, 5204, b"Malassezia sympodialis (strain ATCC 42132)"),
            (1649845, 91347, b"Yersinia pseudotuberculosis complex"),
            (2698737, 2759, b"Sar"),
        ]
    )


class LucaBasedTaxonomyTests(unittest.TestCase):
    def setUp(self):
        taxtab = _get_taxtab()
        self.tax = Taxonomy(taxtab)
        self.nr_species = 30

    def test_root_is_luca(self):
        self.assertEqual(self.tax._get_root_taxon()["Name"], b"LUCA")

    def test_search_luca(self):
        node = self.tax.get_taxnode_from_name_or_taxid("LUCA")
        self.assertEqual(len(node), 1)
        self.assertEqual(node["Name"], b"LUCA")

    def test_extend_genomes(self):
        self.assertEqual(len(self.tax.get_taxid_of_extent_genomes()), self.nr_species)

    def test_search_approx_luca(self):
        self.assertIn("LUCA", [x[1] for x in self.tax.approx_search("luca")])

    def test_numeric_luca_is_found(self):
        self.assertEqual(self.tax.get_taxnode_from_name_or_taxid(0)["Name"], b"LUCA")

    def test_subtaxonomy_does_not_contain_luca_anymore(self):
        # subset of only eukaryotes
        subtax = self.tax.get_induced_taxonomy([451864, 55529, 158441], collapse=True, augment_parents=True)
        self.assertNotIn(b"LUCA", subtax._get_root_taxon()["Name"])

    def test_as_newick_works(self):
        self.assertTrue(self.tax.newick().endswith("LUCA;"), self.tax.newick()[-30:])

    def test_newick_encoder_options(self):
        taxids = [713600, 640131, 37360, 284812, 13735]
        t2s = {
            x: self.tax.tax_table[numpy.where(self.tax.tax_table["NCBITaxonId"] == x)][0]["Name"].decode()
            for x in taxids
        }
        subtax = self.tax.get_induced_taxonomy(taxids, augment_parents=True)
        for quoted in (True, False):
            for leaf in ("sciname", "taxid"):
                for internal in ("taxid", "sciname", None):
                    with self.subTest(quoted=quoted, leaf=leaf, internal=internal):
                        tree = subtax.newick(quoted=quoted, leaf=leaf, internal=internal)
                        exp_leaf = t2s[713600] if leaf == "sciname" else str(713600)
                        exp_int = "2759" if internal == "taxid" else "Eukaryota" if internal == "sciname" else ""
                        if quoted and not leaf == "taxid":
                            exp_leaf = f'"{exp_leaf}"'
                        else:
                            exp_leaf = exp_leaf.replace(" ", "_").replace("(", "[").replace(")", "]")
                        self.assertIn(exp_leaf, tree)
                        self.assertIn(exp_int, tree)


class LucaWithNegTaxIDGenomeTaxonomyTests(LucaBasedTaxonomyTests):
    def setUp(self):
        taxtab = _get_taxtab()
        taxtab = numpy.append(
            taxtab,
            numpy.array([(-2, 451864, b"Some Random Genome", 352)], dtype=taxtab.dtype),
        )
        self.tax = Taxonomy(taxtab)
        self.nr_species = 31

    def test_neg_taxid_species_in_newick(self):
        self.assertIn("Some_Random_Genome", self.tax.newick())


class DBMock(object):
    def __init__(self, h5):
        self.h5 = h5

    def get_hdf5_handle(self):
        return self.h5


class MockDBTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db("TestDb.h5")
        cls.db = DBMock(tables.open_file(path))

    @classmethod
    def tearDownClass(cls):
        cls.db.get_hdf5_handle().close()


class GenomeIdResolverTest(MockDBTestCase):
    def setUp(self):
        self.OmaIdMapper = OmaIdMapper(self.db)

    def test_resolve_uniprot_code(self):
        query = "YEAST"
        g = self.OmaIdMapper.identify_genome(query)
        self.assertEqual(g["UniProtSpeciesCode"], query.encode("ascii"))

    def test_resolve_non_existing_uniprot_code(self):
        query = "XED22"
        with self.assertRaises(UnknownSpecies):
            self.OmaIdMapper.identify_genome(query)

    def test_resolve_taxon_id(self):
        expected = b"PLAF7"
        for query in (36329, b"36329", "36329"):
            g = self.OmaIdMapper.identify_genome(query)
            self.assertEqual(
                g["UniProtSpeciesCode"],
                expected,
                "failed for {} (type {})".format(query, type(query)),
            )

    def test_resolve_nonexisting_code(self):
        with self.assertRaises(UnknownSpecies):
            self.OmaIdMapper.identify_genome(2)

    def test_approx_search_genome(self):
        query = "sacero cervesa"
        expect = "YEAST"
        cands = self.OmaIdMapper.approx_search_genomes(query)
        self.assertIn(expect, [g.uniprot_species_code for g in cands])


class TestPerGenomeMetaData(MockDBTestCase):
    def setUp(self) -> None:
        self.pg = PerGenomeMetaData(self.db.get_hdf5_handle(), "YEAST")

    def test_in_oma_groups_matches(self):
        pass


class FastMapperTester(unittest.TestCase):
    def setUp(self) -> None:
        self.fast_mapper = FastMapper(Database(find_path_to_test_db()))

    def tearDown(self) -> None:
        self.fast_mapper.db.close()

    def test_search_small_sequence(self):
        test_seq = io.StringIO(">test\nAS")
        seq_iter = SeqIO.parse(test_seq, format="fasta")
        with self.assertLogs("pyoma", level=logging.INFO) as cm:
            anno = list(self.fast_mapper.iter_projected_goannotations(seq_iter))
        self.assertIn("Skipping", "\n".join(cm.output))
        self.assertEqual(0, len(anno))

    def check_mapped_seqs(self, seqs, way="Exact"):
        with self.assertLogs("pyoma", level=logging.INFO) as cm:
            anno = list(self.fast_mapper.iter_projected_goannotations(seqs))
        for a in anno:
            with_ = self.fast_mapper.db.id_mapper["OMA"].map_entry_nr(int(a["DB_Object_ID"]))
            self.assertEqual(f"{way}:{with_}".split(":"), a["With"].split(":")[:2])
        self.assertGreaterEqual(len(a), 1)

    def test_search_existing_sequences(self):
        seqs = [SeqRecord(id=f"{enr}", seq=Seq(self.fast_mapper.db.get_sequence(enr).decode())) for enr in range(1, 5)]
        self.check_mapped_seqs(seqs, way="Exact")

    def test_search_inexact_search_seq(self):
        seqs = []
        for en in range(1, 5):
            s = self.fast_mapper.db.get_sequence(en).decode()
            modif_seq = s.replace("A", "R")
            seqs.append(SeqRecord(id=f"{en}", seq=Seq(modif_seq)))
        self.check_mapped_seqs(seqs, way="Approx")



import collections
import unittest
import logging
import time
from unittest.mock import patch

from .test_db import TestWithDbInstance
from pyoma.browser.models import ProteinEntry, HOG
from pyoma.browser.search import (
    OmaGroupSearch,
    HogIDSearch,
    GOSearch,
    TaxSearch,
    SequenceSearch,
    ECSearch,
    DomainSearch,
    XRefSearch,
    SearchResult,
    search,
)

logger = logging.getLogger("search-tests")


class OmaGroupSearchTests(TestWithDbInstance):
    query = 17
    exp_group = 17

    def test_search_grp_nr(self):
        s = OmaGroupSearch(self.db, self.query)
        self.assertEqual([self.exp_group], [g.group_nbr for g in s.search_groups()])

    def test_search_grp_proteins(self):
        s = OmaGroupSearch(self.db, self.query)
        search_res = s.search_entries()
        for p in search_res:
            self.assertEqual(self.exp_group, p.oma_group)


class OmaGroupSearchWithFingerprintTests(OmaGroupSearchTests):
    query = "HAISGRE"


class GOSearchTest(TestWithDbInstance):
    def test_go_term_search(self):
        query = "GO:0000501"
        s = GOSearch(self.db, query)
        res = [e.omaid for e in s.search_entries()]
        self.assertIn("YEAST00011", res)
        self.assertEqual(5, len(res))
        self.assertGreaterEqual(s.count_entries(), len(res))

    def test_go_term_limit(self):
        query = "GO:0000501"
        s = GOSearch(self.db, query)
        s.set_max_entries(2)
        res = [e.omaid for e in s.search_entries()]
        self.assertIn("YEAST00011", res)
        self.assertEqual(2, len(res))
        self.assertGreater(s.count_entries(), len(res))

    def test_go_term_entrynr_filter(self):
        query = "GO:0000501"
        s = GOSearch(self.db, query)
        s.set_entry_nr_filter({1, 2, 11, 102, 103, 3076})
        res = [e.omaid for e in s.search_entries()]
        self.assertIn("YEAST00103", res)
        self.assertEqual(3, len(res))
        self.assertGreater(s.count_entries(), len(res))

    def test_invalid_term_does_list_nothing(self):
        s = GOSearch(self.db, "GO:000000")
        self.assertEqual(0, len(s.search_entries()))


class ECSearchTest(TestWithDbInstance):
    def test_search_full_ec_term(self):
        query = "5.21.5.7"
        with patch.object(self.db, "entrynrs_with_ec_annotation") as mocked:
            mocked.return_value = [12, 4364]  # testdb has no ec annotations
            s = ECSearch(self.db, query)
            self.assertEqual([12, 4364], [p.entry_nr for p in s.search_entries()])
            self.assertEqual(0, s.count_entries())  # check that count is indeed 0


class DomainSearchTest(TestWithDbInstance):
    def test_domain_search_called_with_good_query(self):
        query = b"2.30.29.30"
        with patch.object(self.db.db.get_node("/Annotations/Domains"), "where") as mocked:
            mocked.return_value = [{"EntryNr": 19}]
            s = DomainSearch(self.db, query)
            s.set_entry_nr_filter({5, 511, 19})
            s.search_entries()
            mocked.assert_called_with(
                "(DomainId == dom) & (EntryNr >= 5) & (EntryNr <= 511)",
                condvars={"dom": query},
            )


class TaxSearchTest(TestWithDbInstance):
    def test_existing_tax_of_internal_node(self):
        for query in ("Ascomycota", 4890, "Ascomicotta", "4890"):
            with self.subTest("existing internal node", query=query):
                s = TaxSearch(self.db, query)
                self.assertIn(4890, [z.ncbi_taxon_id for z in s.search_ancestral_genomes()])
                self.assertIn(559292, [z.ncbi_taxon_id for z in s.search_species()])

    def test_existing_extant_node(self):
        for query in ("SCHPO", 284812, "Schizosaccharomyces pombe"):
            with self.subTest("existing extant node", query=query):
                s = TaxSearch(self.db, query)
                self.assertIn(284812, [z.ncbi_taxon_id for z in s.search_species()])

    def test_as_entries_contains_right_ranges(self):
        query = "Ascomycota"
        s = TaxSearch(self.db, query)
        expected_entries = set([])
        for g in s.search_species():
            a, b = self.db.id_mapper["OMA"].genome_range(g.uniprot_species_code)
            expected_entries.update({nr for nr in range(a, b + 1)})
        self.assertSetEqual(expected_entries, s.search_entries())

    def test_from_lowercase_5lettercode(self):
        query = "SchPo"
        s = TaxSearch(self.db, query)
        self.assertIn(284812, [z.ncbi_taxon_id for z in s.search_species()])

    def test_species_only_once_also_after_inexact_match(self):
        query = "sacchar"
        s = TaxSearch(self.db, query)
        res = s.search_species()
        c = collections.Counter((z.ncbi_taxon_id for z in res))
        self.assertEqual(1, max(c.values()))


class SeqSearchTest(TestWithDbInstance):
    def test_search_example(self):
        query = "ADRIAN"
        s = SequenceSearch(self.db, query)
        res = s.search_entries()
        self.assertGreaterEqual(len(res), 1)
        for p in res:
            self.assertIn(query, p.sequence)

    def test_uniprot_formated_seq(self):
        query = """        10         20         30         40         50
MVKETKFYDI LGVPVTATDV EIKKAYRKCA LKYHPDKNPS EEAAEKFKEA
        60         70         80         90        100
SAAYEILSDP EKRDIYDQFG EDGLSGAGGA GGFPGGGFGF GDDIFSQFFG
       110        120        130        140        150
AGGAQRPRGP QRGKDIKHEI SASLEELYKG RTAKLALNKQ ILCKECEGRG
"""
        s = SequenceSearch(self.db, query)
        self.assertEqual(len(s.seq), 150)
        res = s.search_entries()
        self.assertGreaterEqual(len(res), 1)
        for p in res:
            self.assertIn(s.seq.decode(), p.sequence)

    def test_with_existing_fingerprint_as_sequence(self):
        query = "HAISGRE"
        s = SequenceSearch(self.db, query)
        self.assertIn(17, [g.group_nbr for g in s.search_groups()])

    def test_approx_search_contains_extra_info(self):
        query = "EIKKAYRKCKLKYHPAKNPSEEAAEKFKEAAYEILSDP"
        s = SequenceSearch(self.db, query)
        res = s.search_entries()
        self.assertTrue(hasattr(res[0], "alignment"))
        self.assertTrue(hasattr(res[0], "mode"))
        self.assertFalse(hasattr(res[0], "pvaluescore"))


class XRefSearchTest(TestWithDbInstance):
    def test_existing_example(self):
        for query in ("YAS", "K", "YEAST12", "14561"):
            with self.subTest(query=query):
                s = XRefSearch(self.db, query)
                hits = 0
                for p in s.search_entries():
                    for source, val in p.xref_data.items():
                        for el in val:
                            self.assertIn(query.lower(), el.lower())
                            hits += 1
                self.assertGreater(hits, 0)

    def test_total_and_limit(self):
        for query in ("K", "Q1"):
            with self.subTest(query=query):
                s_ref = XRefSearch(self.db, query)
                self.assertGreater(s_ref.estimated_occurrences, 100)
                t0 = time.time()
                res_ref = s_ref.search_entries()
                t1 = time.time()
                s_limit = XRefSearch(self.db, query, max_matches=20)
                res_limit = s_limit.search_entries()
                t2 = time.time()
                self.assertEqual(20, len(res_limit))
                self.assertGreater(len(res_ref), len(res_limit))
                self.assertGreater(t1 - t0 + 2e-3, t2 - t1, "limited search took longer")

    def test_inexisting_term(self):
        for query in ("lksjfewlsd", "Kgop524fslkAA2fnb"):
            with self.subTest(query=query):
                s = XRefSearch(self.db, query)
                hits = 0
                for p in s.search_entries():
                    for source, val in p.xref_data.items():
                        for el in val:
                            self.assertIn(query.lower(), el.lower())
                            hits += 1
                self.assertEqual(hits, 0)


class HogIDSearchTest(TestWithDbInstance):
    def test_existing_hog_with_level(self):
        for query in ("HOG:0000002_4890", "HOG:0000165.1a_4890"):
            with self.subTest(query=query):
                s = HogIDSearch(self.db, query)
                self.assertIn(query.split("_")[0], [h.hog_id for h in s.search_groups()])
                self.assertFalse(s.outdated_query_hog)

    def test_inexact_hogid_with_level(self):
        for query, (exp_hog_id, exp_level, is_root) in zip(
            (
                "HOG:0000002.5a.6c_4890",
                "HOG:0000165.1a_4751",
                "HOG:0000165_4890",
                "HOG:0000165_4751",
            ),
            (
                ("HOG:0000002", "Ascomycota", False),
                ("HOG:0000165.1a", "Ascomycota", True),
                ("HOG:0000165", "Eukaryota", True),
                ("HOG:0000165", "Fungi", False),
            ),
        ):
            with self.subTest(query=query):
                s = HogIDSearch(self.db, query)
                res = s.search_groups()
                self.assertEqual(1, len(res))
                res = res[0]
                self.assertFalse(s.outdated_query_hog)
                self.assertEqual(exp_hog_id, res.hog_id)
                self.assertEqual(exp_level, res.level)
                self.assertEqual(is_root, res.is_root)

    def test_entries_from_hogid(self):
        query = "HOG:0000165.1a"
        s = HogIDSearch(self.db, query)
        self.assertIsNone(s.search_entries())
        tax = TaxSearch(self.db, "Saccharomycetes")
        s.set_taxon_filter(tax.search_ancestral_genomes()[0])
        for e in s.search_entries():
            self.assertIsInstance(e, ProteinEntry)

    def test_bogous_hogid(self):
        query = "HOG:1111111.6a"
        s = HogIDSearch(self.db, query)
        self.assertEqual([], s.search_groups())


class CombineTest(TestWithDbInstance):
    def test_combine_tax_limit(self):
        s1 = TaxSearch(self.db, "Saccharomycetes")
        s2 = SequenceSearch(self.db, "HAISGRE")
        res = SearchResult()
        res &= s1
        self.assertEqual(len(s1.search_entries()), len(res.entries_set))
        self.assertGreater(len(s2.search_entries()), 2)
        res &= s2
        self.assertEqual(len(s2.search_entries()), len(res.entries_set))
        self.assertEqual(2, len(res.species))
        self.assertEqual(1, len(res.ancestral_genomes))
        self.assertEqual(1, len(res.groups))

    def test_combine_tax_and_hog(self):
        # this is a non-furcating species node, so expect to find hogs at level Saccharomycetaceae
        s1 = TaxSearch(self.db, "Saccharomycetes")
        s2 = HogIDSearch(self.db, "HOG:0000165")
        res = SearchResult() & s1 & s2
        self.assertTrue(
            any(
                x.hog_id.startswith("HOG:0000165.1a") and x.level == "Saccharomycetaceae"
                for x in res.groups.values()
                if isinstance(x, HOG)
            ),
            str(res.groups),
        )

    def test_combine_tax_and_xref(self):
        s1 = TaxSearch(self.db, "Saccharomycetes")
        s2 = XRefSearch(self.db, "K")
        unfiltered_xref = s2.search_entries()
        res = SearchResult() & s1 & s2
        self.assertIsNotNone(s2.entry_filter)
        tax_range = s1.search_entries()
        self.assertTrue(all(p.entry_nr in tax_range for p in res.entries.values()))
        self.assertGreater(len(unfiltered_xref), len(res.entries))

    def test_combine_tax_and_go(self):
        s1 = TaxSearch(self.db, "Saccharomycetes")
        s2 = GOSearch(self.db, "GO:0000501")
        res = SearchResult()
        res &= s1
        res &= s2
        self.assertIn(11, [p.entry_nr for p in res.entries.values()])

    def test_combine_tax_and_approx_seq(self):
        s1 = TaxSearch(self.db, "Saccharomycetes")
        s2 = SequenceSearch(self.db, "EIKKAYRKCKLKYHPAKNPSEEAAEKFKEAAYEILSDP")
        res = search([s1, s2])
        self.assertGreaterEqual(len(res.entries), 1)
        for pe in res.entries.values():
            self.assertEqual(pe.mode, "approx")
            self.assertTrue(hasattr(pe, "alignment"))
            self.assertTrue(hasattr(pe, "score"))
        for tax in res.ancestral_genomes.values():
            self.assertTrue(hasattr(tax, "match_score"))
            self.assertLessEqual(tax.match_score, 1)
        for g in res.species.values():
            self.assertTrue(hasattr(g, "match_score"))
            self.assertLessEqual(g.match_score, 0.5)
        self.assertGreaterEqual(s2.count_entries(), len(res.entries))

    def test_combine_xref_and_inexisting_tax(self):
        s1 = TaxSearch(self.db, "HUMAN")
        s2 = XRefSearch(self.db, "K")
        unfiltered_xref = s2.search_entries()
        res = SearchResult() & s1 & s2
        self.assertIsNotNone(s2.entry_filter)
        tax_range = s1.search_entries()
        self.assertEqual(0, len(tax_range))
        self.assertTrue(all(p.entry_nr in tax_range for p in res.entries.values()))
        self.assertGreater(len(unfiltered_xref), len(res.entries))



import itertools
import numpy
import collections
import tables
from .models import ProteinEntry


def chunked(iterator, dtype, n):
    """Batch data into numpy array of length n. The last batch may be shorter."""
    if n < 1:
        raise ValueError("n must be at least one")
    cnt_it, it = itertools.tee((row.fetch_all_fields() for row in iterator), 2)
    while True:
        counter = itertools.count()
        collections.deque(zip(itertools.islice(cnt_it, n), counter), maxlen=0)  # (consume at C speed)
        this_chunk_len = next(counter)
        if this_chunk_len == 0:
            break
        chunk = numpy.fromiter(it, dtype=dtype, count=this_chunk_len)
        yield chunk


def _count_annotations(h5, data):
    ex = numpy.array(
        [
            b"EXP",
            b"IDA",
            b"IPI",
            b"IMP",
            b"IGI",
            b"IEP",
            b"HTP",
            b"HDA",
            b"HMP",
            b"HGI",
            b"HEP",
        ],
        dtype="S3",
    )
    go_tab = h5.get_node("/Annotations/GeneOntology")
    it = go_tab.where('Evidence != b"IEA"')
    NG = len(data)
    for chunk in chunked(it, go_tab.dtype, 10000):
        genome_idx = data["EntryOff"].searchsorted(chunk["EntryNr"]) - 1
        ex_anno = numpy.where(numpy.isin(chunk["Evidence"], ex))[0]
        inc = numpy.bincount(genome_idx[ex_anno], minlength=NG)
        data["NrExpGO"] += inc

        cur_anno = numpy.where(numpy.isin(chunk["Evidence"], ex, invert=True))[0]
        inc = numpy.bincount(genome_idx[cur_anno], minlength=NG)
        data["NrCurGO"] += inc
    for pri, idx in enumerate(numpy.argsort(data, order=("NrExpGO", "NrCurGO"))[::-1]):
        data["Priority"][idx] = pri


class GenomeImportanceHelper:
    dtype = numpy.dtype([("EntryOff", "u4"), ("NrExpGO", "i4"), ("NrCurGO", "i4"), ("Priority", "i4")])
    tab_path = "/Summary/PerGenomeAnnotations"

    def __init__(self, h5: tables.File, compute_stats_if_needed=False):
        try:
            self.genome_data = h5.get_node(self.tab_path).read()
        except tables.NoSuchNodeError:
            self.genome_data = self._get_data(h5, compute_go=compute_stats_if_needed)

    @classmethod
    def create_and_store(cls, h5):
        inst = cls(h5, compute_stats_if_needed=True)
        try:
            tab = h5.get_node(cls.tab_path)
            h5.remove_node(tab)
        except tables.NoSuchNodeError:
            pass
        parent, name = cls.tab_path.rsplit("/", maxsplit=1)
        h5.create_table(
            parent,
            name,
            obj=inst.genome_data,
            createparents=True,
            expectedrows=len(inst.genome_data),
        )
        return inst

    def _get_data(self, h5, compute_go=True):
        offsets = h5.get_node("/Genome").read(field="EntryOff")
        data = numpy.zeros(len(offsets), dtype=self.dtype)
        data["EntryOff"] = offsets
        if compute_go:
            _count_annotations(h5, data)
        else:
            data["Priority"] = numpy.arange(len(data))
        return data

    def sort_numpy_with_entrynr(self, entries: numpy.ndarray):
        """This method sorts a numpy array with entry numbers (either column
        named EntryNr or a single integer array representing the entry numbers)
        according to the importance of the genome for experimentalists.

        :param entries: array containing entry numbers
        :returns: inplace sorted array
        """
        if not isinstance(entries, numpy.ndarray):
            raise ValueError("Invalid type: {}".format(type(entries)))

        if issubclass(entries.dtype.type, numpy.integer):
            col = entries
        elif "EntryNr" in entries.dtype.names:
            col = entries["EntryNr"]
        prios = self.genome_data[self.genome_data["EntryOff"].searchsorted(col) - 1]["Priority"]
        return entries[prios.argsort()]

    def sortkey(self, elements, entrynr_key=None):
        """get a sort key array that can be used to sort the elements
        according to the importance of the corresponding genome.

        :param elements: a list of elements, each corresponding to one protein
        :param entrynr_key: (optinal) a callable that extracts the entry_nr
            from the object. For ProteinModel instances as elements, the method
            automatically accesses the entry_nr using its property.
        """
        default_key = lambda x: x.entry_nr if isinstance(elements[0], ProteinEntry) else lambda x: x
        key = entrynr_key if entrynr_key is not None else default_key
        enrs = numpy.fromiter(map(key, elements), dtype=numpy.int32)
        prios = self.genome_data[self.genome_data["EntryOff"].searchsorted(enrs) - 1]["Priority"]
        return list(prios.argsort())



import logging
import gzip
import bz2
import os
from io import BytesIO, StringIO

package_logger = logging.getLogger("pyoma")
package_logger.addHandler(logging.NullHandler())


def auto_open(fn, *args, **kwargs):
    """function to open regular or compressed files for read / write.

    This function opens files based on their "magic bytes". Supports bz2
    and gzip. If it finds neither of these, presumption is it is a
    standard, uncompressed file.

    Example::

        with auto_open("/path/to/file/maybe/compressed", mode="rb") as fh:
            fh.read()

        with auto_open("/tmp/test.txt.gz", mode="wb") as fh:
            fh.write("my big testfile")

    :param fn: either a string of an existing or new file path, or
        a BytesIO handle
    :param **kwargs: additional arguments that are understood by the
        underlying open handler
    :returns: a file handler
    """
    if isinstance(fn, (BytesIO, StringIO)):
        return fn

    # File opening. This is based on the example on SO here:
    # http://stackoverflow.com/a/26986344
    fmagic = {b"\x1f\x8b\x08": gzip.open, b"\x42\x5a\x68": bz2.open}

    if os.path.isfile(fn) and os.stat(fn).st_size > 0:
        with open(fn, "rb") as fp:
            fs = fp.read(max([len(x) for x in fmagic]))
        for magic, _open in fmagic.items():
            if fs.startswith(magic):
                return _open(fn, *args, **kwargs)
    else:
        if fn.endswith("gz"):
            return gzip.open(fn, *args, **kwargs)
        elif fn.endswith("bz2"):
            return bz2.open(fn, *args, **kwargs)
    return open(fn, *args, **kwargs)



# This Python file uses the following encoding: latin-1
from __future__ import unicode_literals

import tempfile
import unittest

try:
    import unittest.mock as mock
except ImportError:
    import mock as mock
import numpy
import tables
import io
import os
from pyoma.browser import convert as pyoma


@mock.patch.object(
    pyoma.UniProtAdditionalXRefImporter,
    "iter_xreftuples_for_up",
    autospec=True,
    return_value=[],
)
class XRefParsingTest(unittest.TestCase):
    def _create_genome_info(self):
        gs = numpy.zeros(4, dtype=tables.dtype_from_descr(pyoma.tablefmt.GenomeTable))
        gs["EntryOff"] = [0, 2, 3, 4]
        gs["TotEntries"] = [2, 1, 1, 1]
        gs["Release"] = [
            b"Ensembl",
            b"madeup",
            b"Ensembl Metazoa 22; CB4; 13-MAR-2014",
            b"Ensembl Metazoa 27; GCA_000002325.2; 2-JUN-2015",
        ]
        return gs

    def setUp(self):
        self.data = io.StringIO(
            """<E><ID>ENSG00000204640</ID><AC>ENSP00000366061; ENST00000376865</AC><DE>hypotetical protein</DE><GI>125233342</GI><UniProt/TrEMBL>P21522</UniProt/TrEMBL></E>
            <E><ID>BLA22; BLABLA22.Rep22</ID><AC>BLA22.1</AC><EC>3.2.2.-; 4.2.99.18</EC><EntrezGene>32244</EntrezGene><SMR>P21122; Q24S32</SMR><GO>GO:0006270@[[IDA,{'PMID:2167836'}],[IEA,{'GO_REF:002','GO_REF:020','OMA_Fun:001'}]]; GO:0006275@[[IEA,{'GO_REF:002','GO_REF:020','OMA_Fun:001'}]]</GO></E>
            <E><UniProt/TrEMBL>L8ECQ9_BACSU</UniProt/TrEMBL><SwissProt_AC>Q6CI62</SwissProt_AC><SwissProt>ASF1_YARLI</SwissProt><ID>FBgn0218776</ID><AC>FBpp0245919; FBtr0247427</AC><DE>β-hemoglobin</DE><GO></GO></E>
            <E><OS>CAEBR</OS><NR>1</NR><OG>0</OG><AC>CBG23988</AC><CHR>chrI</CHR><ID>CBG23988</ID><LOC>join(80..1057,2068..2664)</LOC><UniProt/TrEMBL>A8WJQ9_CAEBR</UniProt/TrEMBL><GO>GO:0016020@[[IEA,{'GO_REF:038'}]]; GO:0016021@[[IEA,{'GO_REF:038'}]]</GO><SEQ>A</SEQ></E>
            <E><OS>NASVI</OS><NR>11661</NR><OG>0</OG><AC>NV21158-PA; NV21158-RA</AC><CHR>GL340889</CHR><ID>NV21158</ID><LOC>join(113393..113590,114511..114696,114865..114916,115242..115474)</LOC><UniProt/TrEMBL>K7JG62_NASVI</UniProt/TrEMBL><ProtName>Uncharacterized protein</ProtName><SEQ>A</SEQ></E>"""
        )
        self.db_parser = pyoma.DarwinDbEntryParser()
        self.desc_manager = mock.Mock()
        self.go_manager = mock.Mock()
        self.approx_adder = pyoma.ApproximateXRefImporter()
        self.up_extra_adder = pyoma.UniProtAdditionalXRefImporter()
        self.xref_tab = mock.Mock()
        self.ec_tab = mock.Mock()
        self.gs = self._create_genome_info()
        self.importer = pyoma.XRefImporter(
            self.db_parser,
            self.gs,
            self.xref_tab,
            self.ec_tab,
            self.go_manager,
            self.desc_manager,
            self.approx_adder,
            self.up_extra_adder,
        )

    def test_standard_handler(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        enum = pyoma.tablefmt.XRefTable.columns.get("XRefSource").enum
        verif = pyoma.tablefmt.XRefTable.columns.get("Verification").enum
        self.assertIn((1, enum.GI, b"125233342", verif.exact), self.importer.xrefs)
        self.assertIn(
            (3, enum["UniProtKB/SwissProt"], b"ASF1_YARLI", verif.exact),
            self.importer.xrefs,
        )

    def test_multi_handler(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        enum = pyoma.tablefmt.XRefTable.columns.get("XRefSource").enum
        verif = pyoma.tablefmt.XRefTable.columns.get("Verification").enum
        self.assertIn((2, enum["Swiss Model"], b"P21122", verif.unchecked), self.importer.xrefs)
        self.assertIn((2, enum["Swiss Model"], b"Q24S32", verif.unchecked), self.importer.xrefs)

    def test_regex_of_ensembl_ids(self, mock_up):
        for case in ("ENSG00000162687", "ENSMUSP00000162687"):
            match = self.importer.ENS_RE.match(case)
            self.assertIsNotNone(match)

    def test_potential_flush_gets_called(self, mock_up):
        callback = mock.MagicMock(name="potential_flush")
        self.db_parser.add_post_entry_handler(callback)
        self.db_parser.parse_entrytags(self.data)
        self.assertEqual(callback.call_count, 5)

    def test_ensembgenomes_ids(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        enum = pyoma.tablefmt.XRefTable.columns.get("XRefSource").enum
        verif = pyoma.tablefmt.XRefTable.columns.get("Verification").enum
        self.assertIn((4, enum.EnsemblGenomes, b"CBG23988", verif.exact), self.importer.xrefs)

    def test_uniprot_ids(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        enum = pyoma.tablefmt.XRefTable.columns.get("XRefSource").enum
        verif = pyoma.tablefmt.XRefTable.columns.get("Verification").enum
        self.assertIn((3, enum["UniProtKB/TrEMBL"], b"L8ECQ9", verif.exact), self.importer.xrefs)
        self.assertIn((3, enum["UniProtKB/TrEMBL"], b"Q6CI62", verif.exact), self.importer.xrefs)
        mock_up.assert_called_with(self.up_extra_adder, "K7JG62", 5)

    def test_go(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        self.assertEqual(len(self.go_manager.add_annotations.call_args_list), 2)

    def test_ec(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        self.assertIn((2, "3.2.2.-"), self.importer.ec)
        self.assertIn((2, "4.2.99.18"), self.importer.ec)

    def test_nasvi_missing_ac(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        self.importer.flush_buffers()
        enum = pyoma.tablefmt.XRefTable.columns.get("XRefSource").enum
        verif = pyoma.tablefmt.XRefTable.columns.get("Verification").enum
        args, kwargs = self.importer.xref_tab.append.call_args_list[0]
        buffer = args[0]
        self.assertIn(
            (5, enum.SourceAC, b"NV21158-PA", verif.exact),
            buffer,
        )
        self.assertIn(
            (5, enum.SourceAC, b"NV21158-RA", verif.exact),
            buffer,
        )
        self.assertIn(
            (5, enum.EnsemblGenomes, b"NV21158-RA", verif.exact),
            buffer,
        )

    def test_disambiguate(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        enum = pyoma.tablefmt.XRefTable.columns.get("XRefSource").enum
        verif = pyoma.tablefmt.XRefTable.columns.get("Verification").enum
        self.assertIn((3, enum["FlyBase"], b"FBgn0218776", verif.unchecked), self.importer.xrefs)
        self.assertIn((3, enum["FlyBase"], b"FBtr0247427", verif.unchecked), self.importer.xrefs)
        self.assertIn((3, enum["SourceAC"], b"FBtr0247427", verif.exact), self.importer.xrefs)
        self.assertIn((2, enum["SourceID"], b"BLABLA22", verif.exact), self.importer.xrefs)
        self.assertIn((2, enum["SourceID"], b"BLA22", verif.exact), self.importer.xrefs)
        self.assertIn((1, enum["SourceID"], b"ENSG00000204640", verif.exact), self.importer.xrefs)
        self.assertIn(
            (1, enum["Ensembl Gene"], b"ENSG00000204640", verif.exact),
            self.importer.xrefs,
        )
        self.assertIn(
            (1, enum["Ensembl Protein"], b"ENSP00000366061", verif.exact),
            self.importer.xrefs,
        )
        self.assertIn(
            (1, enum["Ensembl Transcript"], b"ENST00000376865", verif.exact),
            self.importer.xrefs,
        )

    def test_descriptions_passed_to_description_manager(self, mock_up):
        self.db_parser.parse_entrytags(self.data)
        self.assertEqual(len(self.desc_manager.add_description.call_args_list), 2)
        self.assertEqual((3, "β-hemoglobin"), self.desc_manager.add_description.call_args[0])

    def test_remove_duplicated_xrefs(self, mock_up):
        ref = (1, 10, "test_id", "unchecked")
        res = (1, 10, b"test_id", 2)
        self.importer.FLUSH_SIZE = 20
        for i in range(self.importer.FLUSH_SIZE + 1):
            self.importer._add_to_xrefs(*ref)
        self.xref_tab.append.assert_not_called()
        self.importer.potential_flush()
        self.xref_tab.append.assert_called_once_with([res])

    def test_remove_near_duplicated_xrefs(self, mock_up):
        refs = [(1, 10, "test_id", "unchecked"), (1, 10, "test_id", "exact")]
        res = (1, 10, b"test_id", 2)
        self.importer.FLUSH_SIZE = 20
        for i in range(self.importer.FLUSH_SIZE + 1):
            for ref in refs:
                self.importer._add_to_xrefs(*ref)
        self.xref_tab.append.assert_not_called()
        self.importer.potential_flush()
        self.xref_tab.append.assert_called_once_with([res])


class DescriptionManagerTest(unittest.TestCase):
    def setUp(self):
        h5file = tables.open_file("test.h5", "w", driver="H5FD_CORE", driver_core_backing_store=0)
        nr_rows = 3
        data = numpy.zeros(nr_rows, dtype=tables.dtype_from_descr(pyoma.tablefmt.ProteinTable))
        data["EntryNr"] = numpy.arange(1, nr_rows + 1)
        h5file.create_table("/", "Entries", pyoma.tablefmt.ProteinTable, obj=data)
        self.h5 = h5file

    def tearDown(self):
        self.h5.close()

    def check_descriptions_one_per_entry(self, descs):
        with pyoma.DescriptionManager(self.h5, "/Entries", "/Descr") as dm:
            for i, desc in enumerate(descs):
                dm.add_description(i + 1, desc)
        exp_lens = numpy.array(list(len(z.encode("utf-8")) for z in descs), dtype="i4")
        exp_offs = numpy.cumsum(exp_lens) - exp_lens[0]
        entry_tab = self.h5.get_node("/Entries")
        numpy.array_equal(exp_lens, entry_tab.col("DescriptionLength"))
        numpy.array_equal(exp_offs, entry_tab.col("DescriptionOffset"))
        txt = self.h5.get_node("/Descr").read().tobytes()
        self.assertEqual("".join(descs).encode("utf-8"), txt)

    def test_add_one_desc_per_entry(self):
        descs = ["hello", "some more lengthy test", "a"]
        self.check_descriptions_one_per_entry(descs)

    def test_add_with_unicode(self):
        descs = ["β-hemoglobin", "", "foobar"]
        self.check_descriptions_one_per_entry(descs)

    def test_severl_descr_per_entry(self):
        descs = [["blabla", "βgamma", "bar"], [""], ["xxx", "yyy"]]
        with pyoma.DescriptionManager(self.h5, "/Entries", "/Descr") as dm:
            for i, per_entry_set in enumerate(descs):
                for desc in per_entry_set:
                    dm.add_description(i + 1, desc)
        exp_txt = "".join(["; ".join(z) for z in descs])
        txt_in_db = self.h5.get_node("/Descr").read().tobytes()
        self.assertEqual(exp_txt.encode("utf-8"), txt_in_db)


class TestGeneOntologyManager(pyoma.GeneOntologyManager):
    ontology_url = "file://" + os.path.dirname(__file__) + "/go-basic-tiny.obo"


class GeneOntologyManagerTest(unittest.TestCase):
    def setUp(self):
        self.h5file = tables.open_file("test.h5", "w", driver="H5FD_CORE", driver_core_backing_store=0)

    def tearDown(self):
        self.h5file.close()

    def store_and_retrieve_data(self, enr, gos):
        with TestGeneOntologyManager(self.h5file, "/Annotation/GO", "/obo") as man:
            man.add_annotations(enr, gos)
        return self.h5file.get_node("/Annotation/GO").read()

    def test_add_none_as_data_raises_ValueError(self):
        with self.assertRaises(ValueError):
            self.store_and_retrieve_data(None, None)

    def test_basic_anno(self):
        res = self.store_and_retrieve_data(
            1,
            "GO:0007610@[[IDA,{'PMID:2167836'}],[IEA,{'GO_REF:002','GO_REF:020','OMA_Fun:001'}]]; GO:0019954@[[IEA,{'GO_REF:002','GO_REF:020','OMA_Fun:001'}]]",
        )
        self.assertEqual(7, len(res))
        numpy.testing.assert_equal(1, res["EntryNr"])
        self.assertEqual(7610, res[0]["TermNr"])
        self.assertEqual(b"IDA", res[0]["Evidence"])
        self.assertEqual(b"PMID:2167836", res[0]["Reference"])

    def test_obo_version_set(self):
        self.store_and_retrieve_data(2, "")
        go_node = self.h5file.get_node("/obo")
        vers = go_node.attrs["ontology_release"]
        self.assertEqual("dummy-test/2017-04-18", vers)

    def test_add_obsolete_term_is_skipped(self):
        import sys

        if sys.version_info >= (3, 4):
            with self.assertLogs("pyoma", level="WARNING") as log:
                res = self.store_and_retrieve_data(1, "GO:0003822@[[IEA,{OMA_Fun:001}]]")
                self.assertIn("invalid GO term for entry 1: GO:0003822", ";".join(log.output))
        else:
            res = self.store_and_retrieve_data(1, "GO:0003822@[[IEA,{OMA_Fun:001}]]")
        self.assertEqual(0, len(res))

    def test_go_obo_version(self):
        self.store_and_retrieve_data(1, "")
        obo = self.h5file.get_node("/obo")
        self.assertEqual("dummy-test/2017-04-18", obo._f_getattr("ontology_release"))


class ApproxXRefImporterTest(unittest.TestCase):
    def setUp(self):
        with tempfile.NamedTemporaryFile(delete=False) as fh:
            self.approx_fname = fh.name
            fh.write(
                """1315\tGeneName\tFca\tmodified
1315\tProtName\tFCA protein\tmodified
1315\tUniProtKB/TrEMBL\tQ6XJS3\tmodified
1406\tORFNames\tAALP_AA5G153500\tmodified
1229\tORFNames\tL914_11153\tmodified""".encode(
                    "utf-8"
                )
            )
        self.approx = pyoma.ApproximateXRefImporter(self.approx_fname)

    def test_non_existing(self):
        self.assertEqual(len(list(self.approx.iter_approx_xrefs_for(10))), 0)

    def test_exitsting(self):
        res = list(self.approx.iter_approx_xrefs_for(1229))
        self.assertEqual(1, len(res), res)
        self.assertEqual(res[0][0], 1229)

    def test_non_sorted_access_raises(self):
        list(self.approx.iter_approx_xrefs_for(100))
        with self.assertRaises(pyoma.InvarianceException):
            list(self.approx.iter_approx_xrefs_for(90))

    def test_return_all_for_large_range(self):
        res = []
        for enr in range(2000):
            res.extend(self.approx.iter_approx_xrefs_for(enr))
        self.assertEqual(len(self.approx.xrefs), len(res))

    def tearDown(self):
        os.remove(self.approx_fname)



from unittest import TestCase
import numpy
from pyoma.browser import xref_contrib


class GeneEntriesTest(TestCase):
    def test_contiguous(self):
        g = xref_contrib.GeneEntries(numpy.arange(7, 12), 8)
        slices = g.entrynr_slices()
        self.assertEqual(1, len(slices))
        self.assertEqual(7, slices[0].start)
        self.assertEqual(12, slices[0].stop)

    def test_gapped(self):
        g = xref_contrib.GeneEntries(numpy.array([1, 3, 4, 6, 7, 8, 10], dtype="i4"), 4)
        slices = g.entrynr_slices()
        self.assertEqual(4, len(slices))
        self.assertEqual(10, slices[3].start)
        self.assertEqual(11, slices[3].stop)


class SpliceHelperTest(TestCase):
    def test_round_robin(self):
        # entry numbers:   1  2  3  4  5  6  7   8  9, 10, 11  12  13
        alt = numpy.array([0, 2, 2, 0, 6, 6, 6, 10, 9, 10, 11, 10, 11], dtype="i4")
        expect_main = [1, 2, 4, 6, 10, 9, 11]
        expect_sets = [{1}, {2, 3}, {4}, {5, 6, 7}, {8, 10, 12}, {9}, {11, 13}]
        slh = xref_contrib.SpliceVariantHelper(None, alt_array=alt)
        genes = list(slh.iter_genes())
        self.assertEqual(len(expect_sets), len(genes))
        for i in range(len(expect_sets)):
            self.assertEqual(expect_main[i], genes[i].main)
            self.assertEqual(expect_sets[i], set(genes[i].enrs))



import os
import unittest
import random
from ..browser.test_db import TestWithDbInstance
import pyoma.application.enrichment


class GOEnrichmentTest(TestWithDbInstance):
    def test_yeast_mapk(self):
        res = pyoma.application.enrichment.extant_species_go_enrichment(self.db, [2834, 678, 1003])
        self.assertIn("GO:0000168", res["GO_ID"].tolist())

    def test_random_set(self):
        sample = random.sample(range(*self.db.id_mapper["OMA"].genome_range(1)), k=15)
        res = pyoma.application.enrichment.extant_species_go_enrichment(self.db, sample)
        self.assertEqual(0, len(res[res["p_bonferroni"] < 0.05]))



from __future__ import unicode_literals, division
from builtins import str
import lxml.etree as etree
import os
import errno
import logging
import re

logger = logging.getLogger(__name__)


class OrthoXMLSplitter(object):
    """Convert orthoxml files with several families.

    This class provides the means to extract a subset of root HOGs (i.e.
    families) into a new output orthoxml file, or to split it and create
    for each family an individual file.

    The object should be instantiated with the input orthoxml file and
    optionally a cache_dir argument where the output orthoxml files will
    be stored. This later parameter can be overwritten in the __call__
    method call that does the work.

    .. note::

       Calls to the splitter will remove the created families from the
       loaded input file, so subsequent calls that contain a family in
       common will miss them from the second call onwards.


    :Example:

      splitter = OrthoXMLSplitter("data.orthoxml", cache_dir="./splits")
      splitter()

    will create files HOGxxxxxx.orthoxml in the ./splits directory."""

    def __init__(self, xml_file, cache_dir=None, release_char=None):
        self.xml_file = xml_file
        if cache_dir is not None:
            self._assert_cache_dir(cache_dir)
        if release_char is None:
            self.release_char = ""
        elif re.match(r"^[A-Z]?$", release_char):
            self.release_char = release_char
        else:
            raise ValueError(
                "unexpected value for release_char: '{}'. Needs to be a single capital ascii letter".format(
                    release_char
                )
            )
        logger.info("loading xml file %s...", xml_file)
        parser = etree.XMLParser(remove_blank_text=True)
        self.Etree_XML = etree.parse(self.xml_file, parser=parser)
        self.Etree_root = self.Etree_XML.getroot()
        logger.info("building lookup table for genes")
        self.gene_lookup = {gene.get("id"): gene for gene in self._iter_gene_elements()}
        logger.info("init of OrthoXMLSplitter finished")

    def _assert_cache_dir(self, cache_dir):
        # Ensure existance of cache directory (py2 compat)
        try:
            os.makedirs(cache_dir)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(cache_dir):
                pass
            else:
                raise
        self.cache_dir = cache_dir

    def _iter_gene_elements(self):
        """This method is a faster version of xpath '//ns:gene'.

        It iterates the element in sequential order"""
        for node in self.Etree_root:
            if node.tag == "{http://orthoXML.org/2011/}species":
                for gene in node.iter("{http://orthoXML.org/2011/}gene"):
                    yield gene

    def _iter_toplevel_groups(self):
        """This method yields all the root hogs sequentially."""
        for node in self.Etree_root:
            if node.tag == "{http://orthoXML.org/2011/}groups":
                for root_hog in node:
                    yield root_hog

    def __call__(
        self,
        hogs_to_extract=None,
        single_hog_files=False,
        basename=None,
        cache_dir=None,
    ):
        """Split/extract hogs from orthoxml file based on root hogs ids.

        Split the input orthoxml or extract a subset of root hogs. If no
        argument is passed, one orthoxml file per root hog is created,
        named as 'HOGxxxxxx.orthoxml', where xxxxxx is the numeric id of
        each hog.

        The set of root hogs to be extracted can be limited by specifying
        a subset of hog ids in the hogs_to_extract parameter. If
        single_hog_files is set to true, each of these hogs will be converted
        into a single orthoxml file named as explained above. If single_hog_files
        is set to false, the whole subset of hogs will be stored in one
        orthoxml file named as specified in `basename`.

        The file(s) will be stored in the cache_dir folder which can be
        specified in the constructor or overwritten as an argument in
        this method.

        :param hogs_to_extract: list or set that contains the set of root
            hogs to be extracted. If set to None, all hogs are extracted.
        :param bool single_hog_files: whether or not to build one orthoxml
            file for all the selected hogs or individual ones.
        :param str basename: name of the output file if a subset of hogs
            is extracted into a single file.
        :param str cache_dir: folder where to store the output files.
        """
        if cache_dir is not None:
            self._assert_cache_dir(cache_dir)
        elif self.cache_dir is None:
            raise RuntimeError("cache dir to output files to is not set")

        if single_hog_files:
            if hogs_to_extract is None:
                raise RuntimeError("useless to extract all hogs into single output file")
            if basename is None or not isinstance(basename, (str, bytes)):
                raise ValueError("basename needs to be specified: {}".format(basename))
            ogs = [og for og in self._iter_toplevel_groups() if int(og.get("id")) in hogs_to_extract]
            fn = os.path.join(self.cache_dir, basename)
            logger.info("extracting %d hogs into %s", len(ogs), fn)
            self.create_new_orthoxml(fn, ogs)
        else:
            for og in self._iter_toplevel_groups():
                if hogs_to_extract is None or int(og.get("id")) in hogs_to_extract:
                    hog_nr = int(og.get("id"))
                    hog_id = "HOG{:07d}.orthoxml".format(hog_nr)
                    fname = os.path.join(self.cache_dir, hog_id)
                    logger.info("extracting %s into %s", hog_id, fname)
                    self.create_new_orthoxml(fname, [og])

    def iter_generefs_in_og(self, og_node):
        for node in og_node.iterdescendants("{http://orthoXML.org/2011/}geneRef"):
            yield node

    def get_gene_via_generef(self, genesref_ids):
        genesref_ids = set(genesref_ids)
        return [self.gene_lookup[gene_id] for gene_id in genesref_ids]

    def create_new_orthoxml(self, fn, OGs):
        """create a new orthoxml file for the passed orthologGroup elements.

        :param fn: the filename of the output file. The path needs to exists
            prior to calling this method.
        :param OGs: the orthologGroup elements that should be included in the
            new output file."""
        # Get element to store
        for og_node in OGs:
            gene_ids = [gene_ref_elem.get("id") for gene_ref_elem in self.iter_generefs_in_og(og_node)]
        gene_els = self.get_gene_via_generef(gene_ids)

        # Get all information to store
        zoo = {}  # <- {key:sp_etree || value: {key:db_el || values:[list_genes]}}
        for gene_el in gene_els:  # <- for all gene el
            db_el = gene_el.getparent().getparent()
            sp_el = db_el.getparent()
            if sp_el in zoo.keys():  # <- if species already visited
                if db_el in zoo[sp_el].keys():  # <- if db already visited so add gene
                    zoo[sp_el][db_el].append(gene_el)
                else:  # <- if db not visited so add db,genes
                    zoo[sp_el][db_el] = []
                    zoo[sp_el][db_el].append(gene_el)
            else:  # <- if species not visited so add sp,db,gene
                zoo[sp_el] = {}
                zoo[sp_el][db_el] = []
                zoo[sp_el][db_el].append(gene_el)

        etree_2_dump = etree.Element("orthoXML", nsmap=self.Etree_root.nsmap)
        for attr, value in self.Etree_root.items():
            etree_2_dump.set(attr, value)

        for species_el in zoo.keys():
            species_xml = etree.Element("species")
            for attr, value in species_el.items():
                species_xml.set(attr, value)
            etree_2_dump.insert(0, species_xml)

            for db_el in zoo[species_el].keys():
                # Add <database> into <species>
                database_xml = etree.SubElement(species_xml, "database")
                for attr, value in db_el.items():
                    database_xml.set(attr, value)

                # Add <genes> TAG into <database>
                genes_xml = etree.SubElement(database_xml, "genes")

                # Fill <genes> with <gene>
                for gene_el in zoo[species_el][db_el]:
                    gene_xml = etree.SubElement(genes_xml, "gene")
                    for attr, value in gene_el.attrib.items():
                        gene_xml.set(attr, value)

        groupsxml = etree.SubElement(etree_2_dump, "groups")
        for og_et in OGs:
            if not og_et.get("id").startswith("HOG:{:s}".format(self.release_char)):
                og_et.set(
                    "id",
                    "HOG:{:s}{:07d}".format(self.release_char, int(og_et.get("id"))),
                )
            groupsxml.append(og_et)

        tree = etree.ElementTree(etree_2_dump)
        tree.write(fn, xml_declaration=True, encoding="utf-8", method="xml", pretty_print=True)



from __future__ import absolute_import, unicode_literals, print_function, division
import unittest
import tempfile
import os
import shutil
import lxml.etree

try:
    import unittest.mock as mock
except ImportError:
    import mock as mock
from pyoma.browser.OrthoXMLSplitter import OrthoXMLSplitter


TEST_ORTHXML = """<?xml version="1.0" encoding="UTF-8"?>
<orthoXML xmlns="http://orthoXML.org/2011/" version="0.3" origin="Test" originVersion="0.2">
<species name="HUMAN" NCBITaxId="9601">
  <database name="HUMANfake" version="0.1">
   <genes>
    <gene id="1" protId="HUMAN1" geneId="HUMANg1" />
    <gene id="2" protId="HUMAN2" geneId="HUMANg2" />
    <gene id="3" protId="HUMAN3" geneId="HUMANg3" />
   </genes>
  </database>
 </species>
 <species name="PANTR" NCBITaxId="9483">
  <database name="PANTRfake" version="0.1">
   <genes>
    <gene id="11" protId="PANTR1" geneId="PANTRg1" />
    <gene id="12" protId="PANTR2" geneId="PANTRg2" />
    <gene id="13" protId="PANTR3" geneId="PANTRg3" />
   </genes>
  </database>
 </species>
 <groups>
  <orthologGroup id="1">
      <property name="TaxRange" value="Primates"/>
      <geneRef id="1" />
      <geneRef id="11" />
  </orthologGroup>
  <orthologGroup id="2">
      <property name="TaxRange" value="Primates"/>
      <geneRef id="2" />
      <geneRef id="12" />
  </orthologGroup>
  <orthologGroup id="3">
      <property name="TaxRange" value="Primates"/>
      <geneRef id="3" />
      <geneRef id="13" />
  </orthologGroup>
 </groups>
</orthoXML>
"""


class OrthoXMLSplitterTester(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.infile = os.path.join(cls.tmpdir, "input.orthoxml")
        with open(cls.infile, "w") as fh:
            fh.write(TEST_ORTHXML)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)

    def setUp(self):
        self.outdir = os.path.join(self.tmpdir, "splits")
        self.splitter = OrthoXMLSplitter(self.infile, self.outdir)
        self.splitter.create_new_orthoxml = mock.MagicMock()

    def test_split_into_single_files(self):
        self.splitter()
        self.assertEqual(3, self.splitter.create_new_orthoxml.call_count)

    def test_extract_subset(self):
        self.splitter(hogs_to_extract=[1, 2])
        self.assertEqual(2, self.splitter.create_new_orthoxml.call_count)

    def test_extract_subset_into_single_file(self):
        self.splitter(hogs_to_extract=[1, 2], single_hog_files=True, basename="single.orthoxml")
        self.assertEqual(1, self.splitter.create_new_orthoxml.call_count)
        args, kwargs = self.splitter.create_new_orthoxml.call_args
        self.assertEqual(args[0], os.path.join(self.outdir, "single.orthoxml"))

    def test_raises_exception_on_single_hog_files_without_filename(self):
        with self.assertRaises(ValueError):
            self.splitter(hogs_to_extract=[1, 2], single_hog_files=True)


class OrthoXMLSplitterResultTester(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.infile = os.path.join(cls.tmpdir, "input.orthoxml")
        with open(cls.infile, "w") as fh:
            fh.write(TEST_ORTHXML)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)

    def setUp(self):
        self.outdir = os.path.join(self.tmpdir, "splits")
        self.splitter = OrthoXMLSplitter(self.infile, self.outdir)()

    def load_data_of_file(self, fn):
        xml = lxml.etree.parse(fn)
        genes = [g.get("id") for g in xml.getroot().findall(".//{http://orthoXML.org/2011/}gene")]
        return genes

    def test_properly_split_in_hogs(self):
        for nr in range(1, 4):
            fn = os.path.join(self.outdir, "HOG{:07d}.orthoxml".format(nr))
            exp = [str(nr), str(nr + 10)]
            self.assertEqual(sorted(self.load_data_of_file(fn)), sorted(exp), "{} failed.".format(fn))



import logging
import random
import unittest

import numpy

from pyoma.browser.db import Database
from pyoma.browser import hogidmap
import tempfile
import os
import math

from .test_db import find_path_to_test_db


class HogIdMapTester(unittest.TestCase):
    db = None

    @classmethod
    def setUpClass(cls):
        path = find_path_to_test_db("TestDb.h5")
        cls.db = Database(path)

    @classmethod
    def tearDownClass(cls):
        cls.db.get_hdf5_handle().close()

    def setUp(self) -> None:
        with tempfile.NamedTemporaryFile(suffix=".h5", delete=False) as fh:
            self.hogmapfn = fh.name

    def tearDown(self) -> None:
        try:
            os.remove(self.hogmapfn)
        except OSError:
            pass

    def yield_hog_hashes_for_family_range(self, start=1, stop=40):
        hasher = hogidmap.HogHasher(self.db)
        for fam in range(start, stop):
            yield from hasher.analyze_fam(fam).items()

    def test_query_exact_identical_groups_will_return_same_hogids(self):
        hash_fam_40 = list(self.yield_hog_hashes_for_family_range(stop=40))
        lsh = hogidmap.LSHBuilder(self.hogmapfn, "w")
        lsh.add_minhashes(hash_fam_40)
        lsh.compute_lsh()
        lsh.close()

        lshq = hogidmap.LSHBuilder(self.hogmapfn, "r")
        for hogid, hashes in hash_fam_40:
            best = sorted(lshq.query(hogid, hashes), key=lambda x: -x[2])[0]
            self.assertEqual(best[0], best[1], "wrong mapping for {}: {}".format(hogid, best[1]))

    def test_stop_and_continue_building_db_works(self):
        hash_fam_40 = list(self.yield_hog_hashes_for_family_range(stop=40))
        lsh = hogidmap.LSHBuilder(self.hogmapfn, "w")
        split = len(hash_fam_40) // 2
        lsh.add_minhashes(hash_fam_40[:split])
        lsh.close()  # close and reopen for append
        lsh = hogidmap.LSHBuilder(self.hogmapfn, "a")
        lsh.add_minhashes(hash_fam_40[split:])
        lsh.compute_lsh()

        for hogid, hashes in hash_fam_40:
            best = sorted(lsh.query(hogid, hashes), key=lambda x: -x[2])[0]
            self.assertEqual(best[0], best[1], "wrong mapping for {}: {}".format(hogid, best[1]))

    def test_inexisting_hog_has_no_match(self):
        lsh = hogidmap.LSHBuilder(self.hogmapfn, "w")
        lsh.add_minhashes(self.yield_hog_hashes_for_family_range(stop=40))
        for hogid, hashes in self.yield_hog_hashes_for_family_range(start=41, stop=44):
            self.assertEqual(0, len(list(lsh.query(hogid, hashes))))

    def test_largely_overlap_matches(self):
        lsh = hogidmap.LSHBuilder(self.hogmapfn, "w")
        lsh.add_minhashes(self.yield_hog_hashes_for_family_range(start=450, stop=480))
        lsh.compute_lsh()

        query_sub_fam = "HOG:0000474.1d.2f"
        all_member_of_subfam = self.db.member_of_hog_id(query_sub_fam)
        nr_matching = math.ceil(len(all_member_of_subfam) * 0.85)
        logging.warning("building samples of {} ids from pool of {}".format(nr_matching, len(all_member_of_subfam)))

        no_match = 0
        for it in range(10):
            query_members = numpy.random.choice(all_member_of_subfam, nr_matching, replace=False)
            minhash = hogidmap.MinHash256()
            for memb in query_members:
                minhash.update(memb["CanonicalId"])
            res = sorted(lsh.query("query", minhash), key=lambda x: -x[2])
            if len(res) == 0:
                no_match += 1
                logging.warning("no match for {}".format(query_members))
                continue
            self.assertEqual(query_sub_fam, res[0][1].decode())
        self.assertLessEqual(no_match, 1)



from __future__ import division, print_function, unicode_literals
import collections
import itertools
import logging
import re

import numpy
import pandas as pd
import tables
from typing import Tuple, Mapping
from .decorators import timethis
from .suffixsearch import SuffixSearcher, SuffixIndexError
from .exceptions import TooUnspecificQuery

logger = logging.getLogger(__name__)


class GeneNamesLookup:
    def __init__(self, h5: tables.File):
        tab = h5.get_node("/XRefIndex/GeneNames").read()
        self.gene_names = {x["XRefId"]: x for x in tab}
        self.lookup_tab = h5.get_node("/XRefIndex/GeneNames_lookup")

    def _low(self, item):
        if isinstance(item, str):
            item = item.encode("utf-8")
        return item.lower()

    def __contains__(self, item):
        return self._low(item) in self.gene_names

    def count(self, term):
        term = self._low(term)
        try:
            v = self.gene_names[term]
            return int(v["Length"])
        except KeyError:
            return 0

    def get_matching_xref_row_nrs(self, term, entrynr_range=None):
        term = self._low(term)
        v = self.gene_names[term]
        lookup = self.lookup_tab[v["Offset"] : v["Offset"] + v["Length"]]
        if entrynr_range is not None:
            lookup = lookup[
                numpy.where(
                    numpy.logical_and(
                        lookup["EntryNr"] >= entrynr_range[0],
                        lookup["EntryNr"] < entrynr_range[1],
                    )
                )
            ]
        return lookup["XRefRow"]


class XRefSearchHelper:
    def __init__(self, h5):
        self.gene_name_lookup = None
        self.reduced_xref_tab = None
        self._re_version = re.compile(r"(?P<base>[\w-]+)\.\d{1,2}$")
        try:
            self.gene_name_lookup = GeneNamesLookup(h5)
            self.reduced_xref_tab = h5.get_node("/XRefIndex/XRef_reduced")
        except tables.NoSuchNodeError:
            logger.warning("No reduced XRef Index and GeneName lookup found")
            pass
        self._use_reduced = self.reduced_xref_tab is not None

        self.xref_tab = h5.get_node("/XRef")
        try:
            self.fulltext_index = SuffixSearcher.from_tablecolumn(self.xref_tab, "XRefId")
        except SuffixIndexError:
            # compability mode
            idx_node = h5.get_node("/XRef_Index")
            self.fulltext_index = SuffixSearcher.from_index_node(idx_node)
        try:
            self._xref_entry_offset = h5.get_node("/XRef_Entry_offset")
        except tables.NoSuchNodeError:
            self._xref_entry_offset = None

    def _version_free_query(self, query: str):
        m = self._re_version.match(query)
        if m is not None:
            return m.group("base")
        return query

    def _query_prefix(self, query: bytes, entrynr_range=None, exact_match=False) -> Tuple[str, Mapping]:
        condvals = {}
        if exact_match:
            stmt = "(XRefId == query)"
            condvals["query"] = query
        else:
            condvals["low"] = query
            condvals["high"] = query[:-1] + bytes([query[-1] + 1])
            stmt = "(XRefId >= low) & (XRefId < high)"
        if entrynr_range is not None:
            stmt += "& (EntryNr >= enr_min) & (EntryNr < enr_max)"
            condvals.update({k: v for k, v in zip(("enr_min", "enr_max"), entrynr_range)})
        return stmt, condvals

    def _prefix_reduced_search(self, query, entrynr_range, limit=None, exact=False):
        def _search(query):
            if query in self.gene_name_lookup:
                return self.gene_name_lookup.get_matching_xref_row_nrs(query, entrynr_range)
            red_tab_it = self.reduced_xref_tab.where(*self._query_prefix(query, entrynr_range, exact))
            xref_rows = numpy.fromiter(itertools.islice((row["XRefRow"] for row in red_tab_it), limit), dtype="i4")
            return xref_rows

        query_lower = query.lower().encode("utf-8")
        xref_rows = _search(query_lower)
        if len(xref_rows) == 0:
            query_lower_without_version = self._version_free_query(query).lower().encode("utf-8")
            if query_lower_without_version != query_lower:
                xref_rows = _search(query_lower_without_version)
        return xref_rows

    def _prefix_reducecd_count(self, query, entrynr_range=None):
        from .db import count_elements

        def _count(query):
            cnts = self.gene_name_lookup.count(query)
            if cnts == 0:
                cnts = count_elements(self.reduced_xref_tab.where(*self._query_prefix(query, entrynr_range)))
            return cnts

        query_lower = query.lower().encode("utf-8")
        count = _count(query_lower)
        if count == 0:
            query_lower_without_version = self._version_free_query(query).lower().encode("utf-8")
            if query_lower_without_version != query_lower:
                count = _count(query_lower_without_version)
        return count

    def _suffix_search(self, query, limit=None, unspecific_exception=50000):
        cnts = self.fulltext_index.count(query)
        if cnts > unspecific_exception:
            raise TooUnspecificQuery(query, cnts)
        conservative_limit = None if limit is None else 4 * limit
        return self.fulltext_index.find(query, limit=conservative_limit)

    def _suffix_count(self, query):
        return self.fulltext_index.count(query)

    def _search_reduced(self, query, mode, entrynr_range, limit):
        exact = mode == "exact"
        if mode in ("prefix", "suffix_if_no_prefix", "exact"):
            xref_rows = self._prefix_reduced_search(query, entrynr_range, limit, exact)
            xref_rows.sort()
        if mode == "suffix" or len(xref_rows) == 0 and mode == "suffix_if_no_prefix":
            xref_rows = self._search_suffix_with_constrains(query, entrynr_range, limit)
        return self.xref_tab[xref_rows]

    def _search_suffix_with_constrains(self, query, entrynr_range, limit):
        args = {}
        if limit is not None:
            unspecific_threshold = max(1000, 10 * limit) if entrynr_range is not None else 50000
            args = {"limit": limit, "unspecific_exception": unspecific_threshold}
        xref_rows = self._suffix_search(query, **args)
        xref_rows.sort()
        if entrynr_range is not None and self._xref_entry_offset is not None:
            row_low = self._xref_entry_offset[entrynr_range[0]]
            row_high = self._xref_entry_offset[entrynr_range[1] + 1]
            xref_rows = xref_rows[numpy.where(numpy.logical_and(xref_rows >= row_low, xref_rows < row_high))]
        return xref_rows

    def _search_direct(self, query, mode, entrynr_range, limit):
        query = query.encode("utf-8")
        exact = mode == "exact"
        if mode in ("prefix", "suffix_if_no_prefix", "exact"):
            print(self.xref_tab)
            print(self.xref_tab.nrows)
            it = self.xref_tab.where(*self._query_prefix(query, entrynr_range, exact_match=exact))
            xrefs = numpy.fromiter(
                (row.fetch_all_fields() for row in itertools.islice(it, limit)),
                dtype=self.xref_tab.dtype,
            )
        if mode == "suffix" or len(xrefs) == 0 and mode == "suffix_if_no_prefix":
            xref_rows = self._search_suffix_with_constrains(query, entrynr_range, limit)
            xrefs = self.xref_tab[xref_rows]
        return xrefs

    def _prefix_direct_count(self, query, entrynr_range=None):
        from .db import count_elements

        query = query.encode("utf-8")
        it = self.xref_tab.where(*self._query_prefix(query, entrynr_range))
        return count_elements(it)

    def search(self, query, mode=None, entrynr_range=None, limit=None):
        if mode is None:
            mode = "suffix_if_no_prefix"
        if mode not in ("suffix", "prefix", "suffix_if_no_prefix", "exact"):
            raise ValueError("invalid search mode: {}".format(mode))
        if self._use_reduced:
            return self._search_reduced(query, mode, entrynr_range, limit)
        else:
            return self._search_direct(query, mode, entrynr_range, limit)

    def count(self, query, mode=None):
        cnts = 0
        if self._use_reduced:
            if mode is None or mode in ("prefix", "suffix_if_no_prefix"):
                cnts = self._prefix_reducecd_count(query)
            if mode == "suffix" or cnts == 0 and mode in (None, "suffix_if_no_prefix"):
                cnts = self._suffix_count(query)
        else:
            if mode is None or mode in ("prefix", "suffix_if_no_prefix"):
                cnts = self._prefix_direct_count(query)
            if mode == "suffix" or cnts == 0 and mode in (None, "suffix_if_no_prefix"):
                cnts = self._suffix_count(query)
        return cnts


class NoSearchXrefIdMapper(object):
    def __init__(self, db, sources=None):
        self._db = db
        self.xref_tab = db.get_hdf5_handle().get_node("/XRef")
        self.xrefEnum = self.xref_tab.get_enum("XRefSource")
        self.idtype = frozenset(nr for src, nr in self.xrefEnum._names.items() if sources is None or src in sources)
        self.verif_enum = self.xref_tab.get_enum("Verification")
        self._max_verif_for_mapping_entrynrs = 1000  # allow all verification values
        try:
            self._xref_entry_offset = db.get_hdf5_handle().get_node("/XRef_EntryNr_offset")
        except tables.NoSuchNodeError:
            self._xref_entry_offset = None

    def map_entry_nr(self, entry_nr):
        """returns the XRef entries associated with the query protein.

        The types of XRefs that are returned depends on the idtype
        class member variable. In the base-class, idtype contains
        all valid xref types. Typically, subclasses of XrefIdMapper
        will change this set.

        :param entry_nr: the numeric id of the query protein.
        :returns: list of dicts with 'source' and 'xref' keys."""
        it = self._iter_xref_for_entry_nr(entry_nr)
        res = [
            {
                "source": self.xrefEnum(row["XRefSource"]),
                "xref": row["XRefId"].decode(),
                "seq_match": self.verif_enum(row["Verification"]),
            }
            for row in it
            if row["XRefSource"] in self.idtype
        ]
        return res

    def _iter_xref_for_entry_nr(self, entry_nr):
        if self._xref_entry_offset is not None:
            start, stop = (self._xref_entry_offset[z] for z in (entry_nr, entry_nr + 1))
            it = self.xref_tab.where(
                "Verification <= {:d}".format(self._max_verif_for_mapping_entrynrs),
                start=start,
                stop=stop,
            )
        else:
            it = self.xref_tab.where(
                "(EntryNr=={:d}) & (Verification <= {:d})".format(entry_nr, self._max_verif_for_mapping_entrynrs)
            )
        return it

    def canonical_source_order(self):
        """returns the list of xref sources in order of their importance.

        Most important source - in the base class for example UniProtKB/SwissProt
        are first. The canonical order is defined in the enum definition.

        :returns: list of source strings"""
        return [self.xrefEnum(z) for z in sorted(self.idtype)]

    def iter_xrefs_for_entry_nr(self, entry_nr):
        """Iterate over the xrefs of a given entry number.

        This method returns a dict with 'source' and 'xref' fields
        (both str) holding the information of the xref record.

        :param entry_nr: the numeric id of the query protein"""
        for row in self._iter_xref_for_entry_nr(entry_nr):
            if row["XRefSource"] in self.idtype:
                yield {
                    "source": self.xrefEnum(row["XRefSource"]),
                    "xref": row["XRefId"].decode(),
                }

    def _combine_query_values(self, field, values):
        parts = ["({}=={})".format(field, z) for z in values]
        return "(" + "|".join(parts) + ")"

    def map_many_entry_nrs(self, entry_nrs):
        """map several entry_nrs with as few db queries as possible
        to their cross-references. The function returns a
        :class:`numpy.recarray` containing all fields as defined in
        the table.

        :param entry_nrs: a list with numeric protein entry ids"""
        mapped_junks = []
        chunk_size = 32
        source_condition = None
        verif_condition = None
        if len(self.idtype) < len(self.xrefEnum):
            chunk_size -= len(self.idtype)  # respect max number of condition variables.
            source_condition = self._combine_query_values("XRefSource", self.idtype)
        if self._max_verif_for_mapping_entrynrs < max(x[1] for x in self.verif_enum):
            chunk_size -= 1
            verif_condition = "(Verification <= {:d})".format(self._max_verif_for_mapping_entrynrs)
        for start in range(0, len(entry_nrs), chunk_size):
            condition_list = [self._combine_query_values("EntryNr", entry_nrs[start : start + chunk_size])]
            if source_condition:
                condition_list.append(source_condition)
            if verif_condition:
                condition_list.append(verif_condition)
            condition = " & ".join(condition_list)
            mapped_junks.append(self.xref_tab.read_where(condition))
        return numpy.lib.recfunctions.stack_arrays(mapped_junks, usemask=False)

    def map_entry_nr_range(self, start, stop):
        """maps for a whole range the entry numbers to the xrefs

        param start: the first entry nr to be mapped
        type start: int, numpy.int
        param stop: the first entry nr that is not included in the result (exlusive)
        type stop: int, numpy.int

        returns: all the mapped xrefs that are respecting
                 the filtering conditions of the actual subtype of
                 XRefIdMapper.
        rtype: :class:`numpy.lib.recarray`"""
        conditions = ["(EntryNr >= {:d}) & (EntryNr < {:d})".format(start, stop)]
        if len(self.idtype) < len(self.xrefEnum):
            source_condition = self._combine_query_values("XRefSource", self.idtype)
            conditions.append(source_condition)
        if self._max_verif_for_mapping_entrynrs < max(x[1] for x in self.verif_enum):
            verif_condition = "(Verification <= {:d})".format(self._max_verif_for_mapping_entrynrs)
            conditions.append(verif_condition)
        query = " & ".join(conditions)
        it = self.xref_tab.where(query)
        try:
            first = next(it)
        except StopIteration:
            return numpy.array([], dtype=self.xref_tab.dtype)
        res = numpy.fromiter(
            map(lambda row: row.fetch_all_fields(), itertools.chain([first], it)),
            dtype=self.xref_tab.dtype,
        )
        return res

    def source_as_string(self, source):
        """string representation of xref source enum value

        this auxiliary method converts the numeric value of
        a xref source into a string representation.

        :param int source: numeric value of xref source"""
        return self.xrefEnum(source)

    def verification_as_string(self, verif):
        """string representation of xref verifiction enum value"""
        return self.verif_enum(verif)

    def xreftab_to_dict(self, tab):
        """convert a xreftable to a dictionary per entry_nr.

        All rows in `tab` are converted into a nested dictionary
        where the outer key is a protein entry number and the
        inner key the xref source type.

        :param tab: a :class:`numpy.recarray` corresponding to XRef
            table definition to be converted"""
        xrefdict = collections.defaultdict(dict)
        for row in tab:
            try:
                typ = self.xrefEnum(row["XRefSource"])
            except IndexError:
                logger.warning("invalid XRefSource value in %s", row)
                continue
            if typ not in xrefdict[row["EntryNr"]]:
                xrefdict[row["EntryNr"]][typ] = {
                    "id": row["XRefId"],
                    "seq_match": self.verif_enum(row["Verification"]),
                }
        return xrefdict


class XrefIdMapper(NoSearchXrefIdMapper):
    def __init__(self, db):
        super(XrefIdMapper, self).__init__(db)
        self.search_helper = XRefSearchHelper(db.get_hdf5_handle())

    @timethis(logging.DEBUG)
    def search_xref(self, xref, is_prefix=False, match_any_substring=False):
        """identify proteins associcated with `xref`.

        The crossreferences are limited to the types in the class
        member `idtype`. In the base class, all types are valid
        xrefs. The method returns a :class:`numpy.recarry` defined
        for the XRef table with all entries pointing to `xref`.

        The method by default returns only exact matches. By setting
        `is_prefix` to True, one can indicated that the requested xref
        should be interpreted as a prefix and all entries matching this
        prefix should be returned.

        :param str xref: an xref to be located
        :param bool is_prefix: treat xref as a prefix and return
                     potentially several matching xrefs
        :param bool match_any_substring: use a suffix index to find
                     any occurrence of the query xref (not limited
                     to the beginning)"""

        if match_any_substring:
            mode = "suffix"
        elif is_prefix:
            mode = "prefix"
        else:
            mode = "exact"
        res = self.search_helper.search(xref, mode=mode)
        if len(res) > 0 and len(self.idtype) < len(self.xrefEnum):
            res = res[numpy.in1d(res["XRefSource"], list(self.idtype))]
        return res

    @timethis(logging.INFO)
    def search_id(self, query, limit=None, entrynr_range=None):
        source_filter = None
        try:
            prefix, term = query.split(":", maxsplit=1)
            if prefix in self.xrefEnum:
                source_filter = self.xrefEnum[prefix]
            else:
                term = query
        except ValueError:
            term = query

        result = collections.defaultdict(dict)
        for xref in self.search_helper.search(term, entrynr_range=entrynr_range, limit=limit):
            if entrynr_range is not None and not (entrynr_range[0] <= xref["EntryNr"] <= entrynr_range[1]):
                continue
            if not source_filter or xref["XRefSource"] == source_filter:
                source = self.xrefEnum(xref["XRefSource"])
                try:
                    result[xref["EntryNr"]][source].append(xref["XRefId"].decode())
                except KeyError:
                    result[xref["EntryNr"]][source] = [xref["XRefId"].decode()]
            if limit is not None and len(result) >= limit:
                break
        return result


class XRefNoApproximateIdMapper(XrefIdMapper):
    def __init__(self, db):
        super(XRefNoApproximateIdMapper, self).__init__(db)
        self._max_verif_for_mapping_entrynrs = self.verif_enum["unchecked"]


class UniProtIdMapper(XrefIdMapper):
    def __init__(self, db):
        super(UniProtIdMapper, self).__init__(db)
        self.idtype = frozenset([self.xrefEnum[z] for z in ["UniProtKB/SwissProt", "UniProtKB/TrEMBL"]])


class LinkoutIdMapper(XrefIdMapper):
    def __init__(self, db):
        super(LinkoutIdMapper, self).__init__(db)
        self.idtype = frozenset(
            [
                self.xrefEnum[z]
                for z in [
                    "UniProtKB/SwissProt",
                    "UniProtKB/TrEMBL",
                    "Ensembl Protein",
                    "Ensembl Gene",
                    "EntrezGene",
                ]
            ]
        )

    def url(self, typ, id_):
        # TODO: improve url generator in external module with all xrefs
        url = None
        try:
            id_ = id_.decode()
        except AttributeError:
            pass

        if typ.startswith("UniProtKB"):
            url = "http://uniprot.org/uniprot/{}".format(id_)
        elif typ == "EntrezGene":
            url = "http://www.ncbi.nlm.nih.gov/gene/{}".format(id_)
        elif typ.startswith("Ensembl"):
            url = "http://ensembl.org/id/{}".format(id_)
        return url

    def xreftab_to_dict(self, tab):
        xref = super(LinkoutIdMapper, self).xreftab_to_dict(tab)
        for d in list(xref.values()):
            for typ, elem in list(d.items()):
                elem["url"] = self.url(typ, elem["id"])
        return xref

    def iter_xrefs_for_entry_nr(self, entry_nr):
        """same as base clase but includes also the url as a field"""
        for xref in super(LinkoutIdMapper, self).iter_xrefs_for_entry_nr(entry_nr):
            xref["url"] = self.url(xref["source"], xref["xref"])
            yield xref


class GeneNameOrSymbolIdMapper(XRefNoApproximateIdMapper):
    def __init__(self, db):
        super(GeneNameOrSymbolIdMapper, self).__init__(db)
        self.order = [
            "Gene Name",
            "UniProtKB/SwissProt",
            "UniProtKB/TrEMBL",
            "HGNC",
            "SourceID",
        ]
        self.idtype = frozenset(self.xrefEnum[z] for z in self.order)

    def canonical_source_order(self):
        return self.order


class GeneNameAndMainDbIdMapper(XRefNoApproximateIdMapper):
    def __init__(self, db):
        super(GeneNameAndMainDbIdMapper, self).__init__(db)
        self.order = [
            "Gene Name",
            "UniProtKB/SwissProt",
            "UniProtKB/TrEMBL",
            "Ensembl Gene",
            "RefSeq",
            "EntrezGene",
            "SourceID",
        ]
        self.idtype = frozenset(self.xrefEnum[z] for z in self.order)

    def canonical_source_order(self):
        return self.order


class DomainNameIdMapper(object):
    def __init__(self, db):
        self.domain_src = db.get_hdf5_handle().root.Annotations.DomainDescription.read()
        self.domain_src.sort(order="DomainId")

    def _get_dominfo(self, domain_id):
        idx = self.domain_src["DomainId"].searchsorted(domain_id)
        if self.domain_src[idx]["DomainId"] != domain_id:
            raise KeyError("no domain info available for {}".format(domain_id))
        return self.domain_src[idx]

    def get_info_dict_from_domainid(self, domain_id):
        info = self._get_dominfo(domain_id)
        return {
            "name": info["Description"].decode(),
            "source": info["Source"].decode(),
            "domainid": domain_id.decode(),
        }
